"""The rm command. Copyright (C) 2023 SynaLinks. License: GPL-3.0"""

from typing import List
from .base import BaseShellCommand
from ...hybridstores.filesystem.context import FileSystemContext
from ...hybridstores.filesystem.filesystem import FileSystem
from ...output_parsers.path import PathOutputParser

class Remove(BaseShellCommand):

    def __init__(self, filesystem: FileSystem):
        super().__init__(
            filesystem = filesystem,
            name = "rm",
            description = "remove the input file or empty folder",
        )
        self.path_parser = PathOutputParser()

    def execute(self, args: List[str], ctx: FileSystemContext) -> str:
        """Method to remove a file or empty folder"""
        if len(args)>0:
            path = args[0]
            path = self.path_parser.parse(path)
            path = ctx.eval_path(path)
        else:
            return "Cannot remove: Missing operand."+\
                " Try 'rm --help' for more information."
        if path.startswith("-"):
            raise ValueError("Cannot remove: Options not supported")
        if not self.filesystem.exists(path):
            raise ValueError(f"Cannot remove {path}: No such file or directory")
        if not self.filesystem.is_empty(path):
            return ValueError(f"Cannot remove {path}: Not empty directory")
        params = {"path": path}
        self.filesystem.hybridstore.query(
            'MATCH (n {name:$path})-[:CONTAINS]->(m) DETACH DELETE m',
            params = params,
        )
        self.filesystem.hybridstore.query(
            'MATCH (n {name:$path}) DETACH DELETE n',
            params = params,
        )
        return f"Sucessfully removed {path}"

    def get_instructions(self) -> str:
        return "The Input should be a unix-like path of the file to be removed."+\
            " Use the option 'rm -r' to remove empty folders"


"""The filesystem context. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

from typing import List
from .path import dirname, join

class FileSystemContext():

    """The filesystem context"""
    def __init__(self):
        self.initialize()

    def eval_path(self, path: str) -> str:
        """Method to eval unix-like path"""
        if path != "":
            if path.startswith(".."):
                path = dirname(self.working_directory) + path[2:]
            elif path.startswith("."):
                path = self.working_directory + path[1:]
            elif path.startswith("~"):
                path = self.home_directory + path[1:]
            elif path == "-":
                path = self.previous_working_directory
            if not path.startswith("/"):
                if self.working_directory == "/":
                    path = self.working_directory + path
                else:
                    path = join([self.working_directory, path])
        else:
            path = self.working_directory
        return path

    def eval_paths(self, paths: List[str]) -> List[str]:
        evaluated_paths = []
        for path in paths:
            evaluated_paths.append(self.eval_path(path))
        return evaluated_paths

    def initialize(self):
        self.home_directory = "/home/user"
        self.working_directory = "/home/user"
        self.previous_working_directory = "/home/user"


from .base import BaseTool, Tool

from .predict import PredictTool

from .duckduckgo_search import DuckDuckGoSearchTool
from .document_search import DocumentSearchTool
from .program_search import ProgramSearchTool
from .past_action_search import PastActionSearchTool
from .entity_search import EntitySearchTool

from .update_objective import UpdateObjectiveTool

from .call_program import CallProgramTool

from .write_file import WriteFileTool
from .append_file import AppendFileTool
from .read_file import ReadFileTool
from .internal_shell import InternalShellTool
from .upload import UploadTool

from .write_program import WriteProgramTool
from .read_program import ReadProgramTool

from .speak import SpeakTool
from .ask_user import AskUserTool

from .clear_trace import ClearTraceTool
from .revert_trace import RevertTraceTool

from .query_facts import QueryFactsTool

from .code_interpreter import CodeInterpreterTool
from .browse_website import BrowseWebsiteTool

__all__ = [
    Tool,
    BaseTool,

    PredictTool,

    DuckDuckGoSearchTool,
    DocumentSearchTool,
    ProgramSearchTool,
    PastActionSearchTool,

    UpdateObjectiveTool,

    CallProgramTool,

    WriteFileTool,
    AppendFileTool,
    ReadFileTool,
    InternalShellTool,
    UploadTool,

    WriteProgramTool,
    ReadProgramTool,

    AskUserTool,
    SpeakTool,

    ClearTraceTool,
    RevertTraceTool,

    QueryFactsTool,

    CodeInterpreterTool,
    BrowseWebsiteTool,
]


from .base import BaseOutputParser
from typing import List

class DecisionOutputParser(BaseOutputParser):
    """
    The output parser for decision making steps
    """

    def parse(self, output: str, options: List[str] = []) -> str:
        toks = output.split()
        if len(toks) > 0:
            token = toks[0].upper().strip(",.")
            if token in options:
                return token
        return output


"""The query facts tool. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import copy
import dspy
from .base import BaseTool
from typing import Optional
from ..hybridstores.fact_memory.fact_memory import FactMemory
from ..output_parsers.cypher import CypherOutputParser
from ..output_parsers.prediction import PredictionOutputParser

class QueryFactsSignature(dspy.Signature):
    """You will be given an objective, purpose and context
    Using the prompt to help you, you will infer the correct Cypher query
    
    Use only the provided relationship types and properties.
    Do not use any other relationship types or properties that are not provided.
    If you cannot generate a Cypher statement based on the provided schema, explain the reason to the user.

    Note: Never give an apology or explain what you are doing."""
    objective = dspy.InputField(desc = "The long-term objective (what you are doing)")
    context = dspy.InputField(desc = "The previous actions (what you have done)")
    purpose = dspy.InputField(desc = "The purpose of the action (what you have to do now)")
    prompt = dspy.InputField(desc = "The action specific instructions (How to do it)")
    cypher_schema = dspy.InputField(desc= "The graph schema")
    cypher_query = dspy.OutputField(desc = "The correct Cypher query")

class QueryFactsTool(BaseTool):

    def __init__(
            self,
            fact_memory: FactMemory,
            lm: Optional[dspy.LM] = None,
        ):
        super().__init__(name = "QueryFacts", lm = lm)
        self.predict = dspy.Predict(QueryFactsSignature)
        self.fact_memory = fact_memory
        self.cypher_parser = CypherOutputParser()
        self.prediction_parser = PredictionOutputParser()

    def query_facts(self, query: str) -> str:
        try:
            result = self.fact_memory.query(query)
            output = [r.values() for r in result.result_set]
            output.insert(0, list(result.keys()))
            return output
        except Exception as err:
            return str(err)
    
    def forward(
            self,
            context: str,
            objective: str,
            purpose: str,
            prompt: str,
            disable_inference: bool = False,
        ) -> dspy.Prediction:
        """Method to perform DSPy forward prediction"""
        if not disable_inference:
            with dspy.context(lm=self.lm if self.lm is not None else dspy.settings.lm):
                pred = self.predict(
                    objective = objective,
                    context = context,
                    purpose = purpose,
                    prompt = prompt,
                    cypher_schema = self.fact_memory.get_schema(refresh=True),
                )
            pred.cypher_query = self.prediction_parser.parse(pred.cypher_query, prefix="\n```cypher", stop=["\n```\n\n"])
            pred.cypher_query = self.cypher_parser.parse(pred.cypher_query)
            output = self.query_facts(pred.query)
            return dspy.Prediction(
                query = pred.query,
                output = output
            )
        else:
            output = self.query_facts(prompt)
            return dspy.Prediction(
                query = prompt,
                output = output
            )

    def __deepcopy__(self, memo):
        cpy = (type)(self)(
            fact_memory = self.fact_memory,
            lm = self.lm,
        )
        cpy.predict = copy.deepcopy(self.predict)
        return cpy


"""The archiver utility. Copyright (C) 2023 SynaLinks. License: GPL-3.0"""

import os
import zipfile
from time import gmtime, strftime
from ..hybridstores.filesystem.path import basename
from ..hybridstores.filesystem.filesystem import FileSystem

class ArchiverUtility():
    """The archiver utility"""

    def __init__(
            self,
            filesystem: FileSystem,
            downloads_directory: str,
        ):
        self.filesystem = filesystem
        self.downloads_directory = downloads_directory

    def zip_and_download(self, path:str) -> str:
        """Method to convert into .zip and download to downloads folder"""
        if not self.filesystem.exists(path):
            raise ValueError("No such file or directory.")
        name = strftime("%Y-%m-%d_%H:%M:%S_", gmtime())+basename(path)
        if not self.downloads_directory:
            self.downloads_directory = os.getcwd()
        filename = os.path.join(self.downloads_directory, name)
        f = zipfile.ZipFile(filename+".zip", mode='a')
        if self.filesystem.is_file(path):
            file_content = self.filesystem.get_document(path)
            f.writestr(basename(path), file_content)
        elif self.filesystem.is_folder(path):
            self.zip_folders_and_files(path, path, f)
        else:
            raise ValueError(f"Cannot upload {path}: Can only upload file or folder")
        f.close()
        return filename+".zip"

    def zip_folders_and_files(
            self,
            target_folder_path:str,
            current_folder_path:str,
            zip_file:zipfile.ZipFile
        ):
        """Method to recursively add folder and files"""
        params = {"folder_path": current_folder_path}
        result_query = self.filesystem.hybridstore.query(
            "MATCH (f:Folder {name:$folder_path})-[:CONTAINS]->(n:Folder) RETURN n",
            params = params,
        )
        for record in result_query.result_set:
            subfolder_path = record[0].properties["name"]
            self.zip_folders_and_files(
                target_folder_path,
                subfolder_path,
                zip_file
            )
        params = {"folder_path": current_folder_path}
        result_query = self.filesystem.hybridstore.query(
            "MATCH (f:Folder {name:$folder_path})-[:CONTAINS]->(n:Document) RETURN n",
            params = params,
        )
        for record in result_query.result_set:
            document_path = record[0].properties["name"]
            file_content = self.filesystem.get_document(document_path)
            zip_file.writestr(
                document_path.replace(target_folder_path, ""),
                file_content,
            )


"""The sentence splitter. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""
# modifed from llama-index
# The MIT License

# Copyright (c) Jerry Liu

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from pydantic import BaseModel
from typing import List, Tuple
from .base import BaseTextSplitter

from .utils import (
    split_by_char,
    split_by_regex,
    split_by_sep,
)

CHUNKING_REGEX = "[^,.;。？！]+[,.;。？！]?"
DEFAULT_PARAGRAPH_SEP = "\n\n"

class Split(BaseModel):
    text: str  # the split text
    is_sentence: bool  # save whether this is a full sentence
    token_size: int  # token length of split text

class SentenceTextSplitter(BaseTextSplitter):

    def __init__(
            self,
            chunk_size: int = 1024,
            chunk_overlap: int = 0,
            separator: str = " ",
            paragraph_separator: str = DEFAULT_PARAGRAPH_SEP,
            secondary_chunking_regex: str = CHUNKING_REGEX,
        ):
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.separator = separator
        self.paragraph_separator = paragraph_separator
        self.secondary_chunking_regex = secondary_chunking_regex

        self._split_fns = [
            split_by_sep(paragraph_separator),
        ]

        self._sub_sentence_split_fns = [
            split_by_regex(secondary_chunking_regex),
            split_by_sep(separator),
            split_by_char(),
        ]

    def split_text(self, text: str) -> List[str]:
        return self._split_text(text, chunk_size=self.chunk_size)

    def _split_text(self, text: str, chunk_size: int) -> List[str]:
        """
        Split incoming text and return chunks with overlap size.

        Has a preference for complete sentences, phrases, and minimal overlap.
        """
        if text == "":
            return [text]
        splits = self._split(text, chunk_size)
        chunks = self._merge(splits, chunk_size)
        return chunks

    def _split(self, text: str, chunk_size: int) -> List[Split]:
        r"""Break text into splits that are smaller than chunk size.

        The order of splitting is:
        1. split by paragraph separator
        2. split by chunking tokenizer (default is nltk sentence tokenizer)
        3. split by second chunking regex (default is "[^,\.;]+[,\.;]?")
        4. split by default separator (" ")

        """
        token_size = self._token_size(text)
        if self._token_size(text) <= chunk_size:
            return [Split(text=text, is_sentence=True, token_size=token_size)]

        text_splits_by_fns, is_sentence = self._get_splits_by_fns(text)

        text_splits = []
        for text_split_by_fns in text_splits_by_fns:
            token_size = self._token_size(text_split_by_fns)
            if token_size <= chunk_size:
                text_splits.append(
                    Split(
                        text=text_split_by_fns,
                        is_sentence=is_sentence,
                        token_size=token_size,
                    )
                )
            else:
                recursive_text_splits = self._split(
                    text_split_by_fns, chunk_size=chunk_size
                )
                text_splits.extend(recursive_text_splits)
        return text_splits

    def _merge(self, splits: List[Split], chunk_size: int) -> List[str]:
        """Merge splits into chunks."""
        chunks: List[str] = []
        cur_chunk: List[Tuple[str, int]] = []  # list of (text, length)
        last_chunk: List[Tuple[str, int]] = []
        cur_chunk_len = 0
        new_chunk = True

        def close_chunk() -> None:
            nonlocal chunks, cur_chunk, last_chunk, cur_chunk_len, new_chunk

            chunks.append("".join([text for text, length in cur_chunk]))
            last_chunk = cur_chunk
            cur_chunk = []
            cur_chunk_len = 0
            new_chunk = True

            # add overlap to the next chunk using the last one first
            # there is a small issue with this logic. If the chunk directly after
            # the overlap is really big, then we could go over the chunk_size, and
            # in theory the correct thing to do would be to remove some/all of the
            # overlap. However, it would complicate the logic further without
            # much real world benefit, so it's not implemented now.
            if len(last_chunk) > 0:
                last_index = len(last_chunk) - 1
                while (
                    last_index >= 0
                    and cur_chunk_len + last_chunk[last_index][1] <= self.chunk_overlap
                ):
                    text, length = last_chunk[last_index]
                    cur_chunk_len += length
                    cur_chunk.insert(0, (text, length))
                    last_index -= 1

        while len(splits) > 0:
            cur_split = splits[0]
            if cur_split.token_size > chunk_size:
                raise ValueError("Single token exceeded chunk size")
            if cur_chunk_len + cur_split.token_size > chunk_size and not new_chunk:
                # if adding split to current chunk exceeds chunk size: close out chunk
                close_chunk()
            else:
                if (
                    cur_split.is_sentence
                    or cur_chunk_len + cur_split.token_size <= chunk_size
                    or new_chunk  # new chunk, always add at least one split
                ):
                    # add split to chunk
                    cur_chunk_len += cur_split.token_size
                    cur_chunk.append((cur_split.text, cur_split.token_size))
                    splits.pop(0)
                    new_chunk = False
                else:
                    # close out chunk
                    close_chunk()

        # handle the last chunk
        if not new_chunk:
            chunk = "".join([text for text, length in cur_chunk])
            chunks.append(chunk)

        # run postprocessing to remove blank spaces
        return self._postprocess_chunks(chunks)

    def _postprocess_chunks(self, chunks: List[str]) -> List[str]:
        """Post-process chunks.
        Remove whitespace only chunks and remove leading and trailing whitespace.
        """
        new_chunks = []
        for chunk in chunks:
            stripped_chunk = chunk.strip()
            if stripped_chunk == "":
                continue
            new_chunks.append(stripped_chunk)
        return new_chunks

    def _token_size(self, text: str) -> int:
        return len(text)

    def _get_splits_by_fns(self, text: str) -> Tuple[List[str], bool]:
        for split_fn in self._split_fns:
            splits = split_fn(text)
            if len(splits) > 1:
                return splits, True

        for split_fn in self._sub_sentence_split_fns:
            splits = split_fn(text)
            if len(splits) > 1:
                break

        return splits, False


"""The read file tool. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import copy
import dspy
from .base import BaseTool
from typing import Optional
from ..hybridstores.filesystem.filesystem import FileSystem
from ..utility.reader import ReaderUtility
from ..output_parsers.path import PathOutputParser
from ..output_parsers.prediction import PredictionOutputParser
from ..types.state import AgentState

class ReadFileSignature(dspy.Signature):
    """You will be given an objective, purpose and context
    Using the prompt to help you, you will infer the correct filename
    
    Note: Never give an apology or explain what you are doing."""
    objective = dspy.InputField(desc = "The long-term objective (what you are doing)")
    context = dspy.InputField(desc = "The previous actions (what you have done)")
    purpose = dspy.InputField(desc = "The purpose of the action (what you have to do now)")
    prompt = dspy.InputField(desc = "The action specific instructions (How to do it)")
    filename = dspy.OutputField(desc = "The name of the file to read")

class ReadFileTool(BaseTool):

    def __init__(
            self,
            filesystem: FileSystem,
            agent_state: AgentState,
            lm: Optional[dspy.LM] = None,
        ):
        super().__init__(name = "ReadFile", lm = lm)
        self.predict = dspy.Predict(ReadFileSignature)
        self.agent_state = agent_state
        self.filesystem = filesystem
        self.reader = ReaderUtility(filesystem=self.filesystem)
        self.prediction_parser = PredictionOutputParser()
        self.path_parser = PathOutputParser()

    def read_file(self, filename: str) -> str:
        try:
            filename = self.agent_state.context.eval_path(filename)
            return self.reader.read_document(filename)
        except Exception as err:
            return str(err)
    
    def forward(
            self,
            context: str,
            objective: str,
            purpose: str,
            prompt: str,
            disable_inference: bool = False,
        ) -> dspy.Prediction:
        """Method to perform DSPy forward prediction"""
        if not disable_inference:
            pred = self.predict(
                objective = objective,
                context = context,
                purpose = purpose,
                prompt = prompt,
            )
            pred.filename = self.prediction_parser.parse(
                pred.filename,
                prefix="Filename:",
                stop = ["\n"],
            )
            pred.filename = self.path_parser.parse(pred.filename)
            dspy.Suggest(
                len(pred.filename) < 250,
                "Filename must be short and consice"
            )
            observation = self.read_file(pred.filename)
            return dspy.Prediction(
                filename = pred.filename,
                content = observation,
            )
        else:
            observation = self.read_file(prompt)
            return dspy.Prediction(
                filename = prompt,
                content = observation,
            )

    def __deepcopy__(self, memo):
        cpy = (type)(self)(
            filesystem = self.filesystem,
            agent_state = self.agent_state,
            lm = self.lm,
        )
        cpy.predict = copy.deepcopy(self.predict)
        return cpy


"""The path output parser. Copyright (C) 2023 SynaLinks. License: GPL-3.0"""

from .base import BaseOutputParser

class PathOutputParser(BaseOutputParser):
    """
    The Output Parser for the paths
    """
    def parse(self, output:str) -> str:
        """Fix and validate the given path."""
        path = output.strip()
        path = path.replace("\\_", "_")
        path = path.strip("\"'")
        path = path.rstrip(",.;")
        if len(path) > 1:
            path = path.rstrip("/")
        return path


from typing import Optional
from .base import BaseKnowledgeParser
from ..hybridstores.fact_memory.fact_memory import FactMemory
from ..hybridstores.filesystem.filesystem import FileSystem

class TextKnowledgeParser(BaseKnowledgeParser):

    def __init__(
            self,
            filesystem: FileSystem,
            fact_memory: FactMemory,
        ):
        super().__init__(
            filesystem = filesystem,
            fact_memory = fact_memory,
            valid_extensions = [".txt", ".md"],
        )

    def parse(self, dest_path:str, content: str):
        self.filesystem.add_texts(
            texts=[content],
            ids=[dest_path],
        )
        # TODO add triplet extraction for the fact memory

    def read(self, source_path: str) -> str:
        f = open(source_path, "r")
        file_content = f.read()
        return file_content


"""The fact memory. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""
# modifed from llama-index to add vector embeddings
# The MIT License

# Copyright (c) Jerry Liu

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from ..hybridstore import HybridStore
from typing import List, Dict, Optional, Any
from ..hybridstore import HybridStore
from ...embeddings.base import BaseEmbeddings

class FactMemory(HybridStore):
    """The fact memory"""

    def __init__(
        self,
        index_name: str,
        embeddings: BaseEmbeddings,
        graph_index: str = "fact_memory",
        hostname: str = "localhost",
        port: int = 6379,
        username: str = "",
        password: str = "",
        indexed_label: str = "Entity",
        wipe_on_start: bool = False,
        chunk_size: int = 1024,
        chunk_overlap: int = 0,
    ):
        super().__init__(
            index_name = index_name,
            graph_index = graph_index,
            embeddings = embeddings,
            hostname = hostname,
            port = port,
            username = username,
            password = password,
            indexed_label = indexed_label,
            wipe_on_start = wipe_on_start,
        )
        self.schema = ""

    def get_rel_map(
        self, subjs: Optional[List[str]] = None, depth: int = 2, limit: int = 30
    ) -> Dict[str, List[List[str]]]:
        """Get flat rel map."""
        # The flat means for multi-hop relation path, we could get
        # knowledge like: subj -> rel -> obj -> rel -> obj -> rel -> obj.
        # This type of knowledge is useful for some tasks.
        # +-------------+------------------------------------+
        # | subj        | flattened_rels                     |
        # +-------------+------------------------------------+
        # | "player101" | [95, "player125", 2002, "team204"] |
        # | "player100" | [1997, "team204"]                  |
        # ...
        # +-------------+------------------------------------+

        rel_map: Dict[Any, List[Any]] = {}
        if subjs is None or len(subjs) == 0:
            # unlike simple graph_store, we don't do get_all here
            return rel_map

        query = f"""
            MATCH (n1:{self.indexed_label})
            WHERE n1.name IN $subjs
            WITH n1
            MATCH p=(n1)-[e*1..{depth}]->(z)
            RETURN p LIMIT {limit}
        """

        data = self.hybridstore.query(query, params={"subjs": subjs})
        if not data.result_set:
            return rel_map

        for record in data.result_set:
            nodes = record[0].nodes()
            edges = record[0].edges()

            subj_id = nodes[0].properties["name"]
            path = []
            for i, edge in enumerate(edges):
                dest = nodes[i + 1]
                dest_id = dest.properties["name"]
                path.append(edge.relation)
                path.append(dest_id)

            paths = rel_map[subj_id] if subj_id in rel_map else []
            paths.append(path)
            rel_map[subj_id] = paths

        return rel_map

    def refresh_schema(self) -> None:
        """Refreshes the graph schema information."""
        node_properties = self.hybridstore.query("CALL db.PROPERTYKEYS()")
        relationships = self.hybridstore.query("CALL db.RELATIONSHIPTYPES()")

        self.schema = f"""
Properties: {node_properties.result_set}
Relationships: {relationships.result_set}
"""

    def get_schema(self, refresh: bool = True) -> str:
        """Get the schema of the hybridstore."""
        if self.schema and not refresh:
            return self.schema
        self.refresh_schema()
        return self.schema

    def get_triplets(self, subj: str) -> List[List[str]]:
        """Get triplets."""
        get_query = f"""
            MATCH (n1:`{self.indexed_label}`)-[r]->(n2:`{self.indexed_label}`)
            WHERE n1.id = $subj RETURN type(r), n2.id
        """
        result = self.hybridstore.query(
            get_query, params={"subj": subj}
        )
        return result.result_set

    def add_triplet(self, subj: str, rel: str, obj: str) -> None:
        """Add triplet."""

        query = """
            MERGE (n1:`%s` {name:$subj})
            MERGE (n2:`%s` {name:$obj})
            MERGE (n1)-[:`%s`]->(n2)
        """

        prepared_statement = query % (
            self.indexed_label,
            self.indexed_label,
            rel.replace(" ", "_").upper(),
        )

        # Call with prepared statement
        self.hybridstore.query(prepared_statement, params={"subj": subj, "obj": obj})

    def delete_triplet(self, subj: str, rel: str, obj: str) -> None:
        """Delete triplet."""

        def delete_rel(subj: str, obj: str, rel: str) -> None:
            rel = rel.replace(" ", "_").upper()
            query = f"""
                MATCH (n1:`{self.indexed_label}`)-[r:`{rel}`]->(n2:`{self.indexed_label}`)
                WHERE n1.name = $subj AND n2.name = $obj DELETE r
            """
            self.hybridstore.query(query, params={"subj": subj, "obj": obj})

        def delete_entity(entity: str) -> None:
            query = f"MATCH (n:`{self.indexed_label}`) WHERE n.name = $entity DELETE n"
            self.hybridstore.query(query, params={"entity": entity})

        def check_edges(entity: str) -> bool:
            query = f"""
                MATCH (n1:`{self.indexed_label}`)--()
                WHERE n1.name = $entity RETURN count(*)
            """
            result = self.hybridstore.query(
                query, params={"entity": entity}
            )
            return bool(result.result_set)

        delete_rel(subj, obj, rel)
        if not check_edges(subj):
            delete_entity(subj)
        if not check_edges(obj):
            delete_entity(obj)


"""The mv command. Copyright (C) 2023 SynaLinks. License: GPL-3.0"""

from typing import List
from .base import BaseShellCommand
from ...hybridstores.filesystem.context import FileSystemContext
from ...hybridstores.filesystem.filesystem import FileSystem
from ...hybridstores.filesystem.path import dirname
from ...output_parsers.path import PathOutputParser

class Move(BaseShellCommand):
    path_parser: PathOutputParser = PathOutputParser()

    def __init__(self, filesystem: FileSystem):
        super().__init__(
            filesystem = filesystem,
            name = "mv",
            description = "move the target file or folder to destination",
        )

    def execute(self, args: List[str], ctx: FileSystemContext) -> str:
        """Method to move a file or folder"""
        target_path = ""
        destination_path = ""
        if len(args) < 2:
            raise ValueError(
                "Cannot move: Missing file operand."+
                " Try 'rm --help' for more information."
            )
        target_path = args[0]
        target_path = self.path_parser.parse(target_path)
        destination_path = args[1]
        destination_path = self.path_parser.parse(destination_path)
               
        target_path = ctx.eval_path(target_path)
        destination_path = ctx.eval_path(destination_path)

        if not self.filesystem.exists(target_path):
            raise ValueError(f"Cannot move: No such file")
        if self.filesystem.exists(destination_path):
            raise ValueError("Cannot move: File or directory already exists")

        if dirname(target_path) == dirname(destination_path):
            # If same directory, just rename the file
            params = {"target": target_path, "destination": destination_path}
            self.filesystem.hybridstore.query(
                'MATCH (n {name:$target}) SET n.name=$destination',
                params = params,
            )
        else:
            # Otherwise remove the Folder's edge and recreate it + rename
            params = {"target": target_path}
            self.filesystem.hybridstore.query(
                'MATCH (n {name:$target})<-[r:CONTAINS]-(m:Folder) DELETE r',
                params = params,
            )
            folder_path = dirname(destination_path)
            params = {"folder_path": folder_path, "target": target_path}
            self.filesystem.hybridstore.query(
                'MATCH (n:Folder {name:$folder_path}),'+
                ' (m {name:$target}) MERGE (n)-[:CONTAINS]->(m)',
                params = params,
            )
            params = {"target": target_path, "destination": destination_path}
            self.filesystem.hybridstore.query(
                'MATCH (n {name:$target}) SET n.name=$destination',
                params = params,
            )
        return "Sucessfully moved"

    def get_instructions(self) -> str:
        return "The Input should be the unix-like target path"+\
            " then the unix-like destination path."




"""The write file tool. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import copy
import dspy
from .base import BaseTool
from typing import Optional
from ..hybridstores.program_memory.program_memory import ProgramMemory
from ..output_parsers.program_name import ProgramNameOutputParser
from ..output_parsers.cypher import CypherOutputParser
from ..output_parsers.prediction import PredictionOutputParser
from ..utility.tester import TesterUtility

class WriteProgramSignature(dspy.Signature):
    """You will be given an objective, purpose and context
    Using the prompt to help you, you will infer the correct filename and cypher query
    
    Note: Never give an apology or explain what you are doing."""
    objective = dspy.InputField(desc = "The long-term objective (what you are doing)")
    context = dspy.InputField(desc = "The previous actions (what you have done)")
    purpose = dspy.InputField(desc = "The purpose of the action (what you have to do now)")
    prompt = dspy.InputField(desc = "The action specific instructions (How to do it)")
    filename = dspy.OutputField(desc="The name of the .cypher file (short and concise)")
    cypher_query = dspy.OutputField(desc = "The Cypher query to write into memory (make sure to comment non-Cypher lines)")

class WriteProgramTool(BaseTool):

    def __init__(
            self,
            program_memory: ProgramMemory,
            lm: Optional[dspy.LM] = None,
        ):
        super().__init__(name = "WriteProgram", lm = lm)
        self.predict = dspy.Predict(WriteProgramSignature)
        self.program_memory = program_memory
        self.program_name_parser = ProgramNameOutputParser()
        self.cypher_parser = CypherOutputParser()
        self.prediction_parser = PredictionOutputParser()
        self.program_tester = TesterUtility(program_memory=program_memory)

    def write_program(self, filename: str, content: str) -> str:
        try:
            self.program_tester.verify_programs(
                [filename],
                [content],
            )
            self.program_memory.add_texts(texts = [content], ids = [filename])
            return "Successfully written"
        except Exception as err:
            return str(err)
    
    def forward(
            self,
            context: str,
            objective: str,
            purpose: str,
            prompt: str,
            disable_inference: bool = False,
        ) -> dspy.Prediction:
        """Method to perform DSPy forward prediction"""
        if not disable_inference:
            with dspy.context(lm=self.lm if self.lm is not None else dspy.settings.lm):
                pred = self.predict(
                    objective = objective,
                    context = context,
                    purpose = purpose,
                    prompt = prompt,
                )
            pred.filename = self.prediction_parser.parse(pred.filename, prefix="File Name:", stop=["\n"])
            pred.filename = self.program_name_parser.parse(pred.filename)
            pred.cypher_query = self.prediction_parser.parse(pred.cypher_query, prefix="\n```cypher", stop=["\n```\n\n"])
            pred.cypher_query = self.cypher_parser.parse(pred.cypher_query)
            dspy.Suggest(
                len(pred.filename) != 0,
                "Filename should not be empty"
            )
            dspy.Suggest(
                len(pred.filename) < 250,
                "Filename should be short and consice"
            )
            observation = self.write_program(pred.filename, pred.cypher_query)
            return dspy.Prediction(
                filename = pred.filename,
                content = pred.cypher_query,
                observation = observation,
            )
        else:
            raise NotImplementedError("Disabling inference for WriteProgram not supported")

    def __deepcopy__(self, memo):
        cpy = (type)(self)(
            program_memory = self.program_memory,
            lm = self.lm,
        )
        cpy.predict = copy.deepcopy(self.predict)
        return cpy


import dspy
from pydantic import BaseModel

class CritiqueTrace(dspy.Signature):
    """Critique the trace based on the objective to assess"""
    objective = dspy.InputField(desc="The objective")
    trace = dspy.InputField(desc="The trace to assess")
    critique = dspy.OutputField(desc="The critique of the trace")

class Score(BaseModel):
    score: float

class CritiqueToScoreSignature(dspy.Signature):
    """Convert a critique into a score between 0.0 (bad) and 1.0 (good)"""
    critique = dspy.InputField(desc="The critique to convert into a score")
    score: Score = dspy.OutputField(desc="A score between 0.0 and 1.0")

def critique_trace(example, prediction, trace=None):
    # This line means that we discard the example if the agent reached the max iterations
    # Meaning it was probably stuck in a loop
    if prediction.finish_reason == "max iters":
        return False
    pred = dspy.ChainOfThought(CritiqueTrace)(
        objective = example.objective,
        trace = prediction.program_trace,
    )
    result = dspy.TypedPredictor(CritiqueToScoreSignature)(
        critique = pred.critique,
    )
    return result.score.score


"""The file output parser. Copyright (C) 2023 SynaLinks. License: GPL-3.0"""

import re
from typing import Tuple, List
from .base import BaseOutputParser

PATTERN = \
r"(?P<filename>[a-zA-Z0-9_.-]+)\n```(?P<lang>[a-zA-Z]+)\n(?P<content>.*?)\n```"

class FileOutputParser(BaseOutputParser):
    """
    The Output Parser for the files
    """
    def parse(self, output:str) -> List[Tuple[str, str, str]]:
        """Fix and validate the output."""
        filenames = []
        contents = []
        languages = []
        
        output = output.strip()
        matches = re.finditer(PATTERN, output, re.DOTALL)

        for match in matches:
            filename = match.group('filename').replace("\\_", "_")
            filename = filename.strip(".'\"")
            filenames.append(filename)
            contents.append(match.group('content').strip())
            languages.append(match.group('lang'))

        if len(filenames) == 0:
            raise ValueError(
                "Error while parsing the Input: Invalid format."\
                +" Please make sure that the file has a name and the format is correct."
            )
        return filenames, contents, languages


"""The hybridstore. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import numpy as np
import uuid
import json
from falkordb import FalkorDB, Graph
from typing import List, Optional, Dict, Any
from ..embeddings.base import BaseEmbeddings

RESERVED_NAMES = [
    "main",
    "playground",
    "filesystem",
    "program_memory",
    "trace_memory",
    "fact_memory",
    "\u006D\u0061\u0069\u006E",
    "\u0070\u006C\u0061\u0079\u0067\u0072\u006F\u0075\u006E\u0064",
    "\u0066\u0069\u006C\u0065\u0073\u0079\u0073\u0074\u0065\u006D",
    "\u0070\u0072\u006F\u0067\u0072\u0061\u006D\u005F\u006D\u0065\u006D\u006F\u0072\u0079",
    "\u0074\u0072\u0061\u0063\u0065\u005F\u006D\u0065\u006D\u006F\u0072\u0079",
    "\u0066\u0061\u0063\u0074\u005f\u006d\u0065\u006d\u006f\u0072\u0079"
]

class HybridStore():
    """The base class for hybridstores"""

    def __init__(
            self,
            index_name: str,
            graph_index: str,
            embeddings: BaseEmbeddings,
            hostname: int = "localhost",
            port: int = 6379,
            username: str = "",
            password: str = "",
            indexed_label: str = "Content",
            wipe_on_start: bool = False,
        ):
        self.hostname = hostname
        self.port = port
        self.username = username
        self.password = password
        self.index_name = index_name
        self.graph_index = graph_index
        self.embeddings = embeddings
        self.indexed_label = indexed_label
        self.wipe_on_start = wipe_on_start
        self.client = FalkorDB(
            hostname,
            port,
            username = username if username else None,
            password = password if password else None,
        )
        self.hybridstore = self.get_graph(self.graph_index)
        if self.wipe_on_start:
            self.clear()
        self.init_index()

    def add_texts(
            self,
            texts: List[str],
            ids: List[str] = [],
            descriptions: List[str] = [],
            metadatas: List[Dict[str, str]] = [],
        ) -> List[str]:
        """Method to add texts"""
        indexes = []
        if ids:
            assert(len(texts) == len(ids))
        if metadatas:
            assert(len(texts) == len(metadatas))
        for idx, text in enumerate(texts):
            content_index = str(uuid.uuid4().hex) if not ids else ids[idx]
            description = text if not descriptions else descriptions[idx]
            metadata = {} if not metadatas else metadatas[idx]
            vector = self.embeddings.embed_text(description)
            params = {
                "vector": list(vector),
                "index": content_index,
                "metadata": json.dumps(metadata, indent=2),
                "description": description,
                "content": text,
            }
            if self.exists(content_index):
                self.hybridstore.query(
                    "MATCH (n:"+self.indexed_label+" {name:$index}) SET "+
                    "n.embeddings_vector=vecf32($vector), "+
                    "n.metadata=$metadata, n.description=$description, n.content=$content",
                    params,
                )
            else:
                self.hybridstore.query(
                    "MERGE (n:"+self.indexed_label+" {name:$index, "+
                    "embeddings_vector:vecf32($vector), "+
                    "metadata:$metadata, description:$description, content:$content})",
                    params,
                )
            indexes.append(content_index)
        return indexes

    def get_graph(self, graph_index: str) -> Graph:
        """Retreive a graph from the KB"""
        return self.client.select_graph(self.index_name+":graph:"+graph_index)

    def exists(self, index:str, label:str = "") -> bool:
        """Method to check if an entry is present in the hybridstore"""
        params = {"index": index}
        if label:
            result = self.hybridstore.query(
                'MATCH (n:'+label+' {name:$index}) RETURN n',
                params = params,
            )
        else:
            result = self.hybridstore.query(
                'MATCH (n {name:$index}) RETURN n',
                params = params,
            )
        if len(result.result_set) > 0:
            return True
        return False

    def remove_texts(self, indexes: List[str]):
        """Method to remove texts"""
        for idx in indexes:
            self.delete_content(idx)

    def set_content(self, content_index: str, text: str) -> bool:
        """Set content into FalkorDB"""
        if not self.exists(content_index):
            params = {"index": content_index, "content": text}
            self.hybridstore.query(
                'MERGE (n:'+self.indexed_label+' {name:$index, content:$content})',
                params = params,
            )
        else:
            params = {"index": content_index, "content": text}
            self.hybridstore.query(
                'MATCH (n:'+self.indexed_label+' {name:$index}) SET n.content=$content',
                params = params,
            )
        return True

    def get_content(self, content_index: str) -> str:
        """Get content from FalkorDB"""
        if self.exists(content_index):
            params = {"index": content_index}
            result = self.hybridstore.query(
                'MATCH (n:Content {name:$index}) RETURN n.content AS content',
                params=params,
            )
            if len(result.result_set) > 0:
                return result.result_set[0][0]
        return ""

    def set_content_description(
            self,
            content_index: str,
            description: str,
        ) -> bool:
        """Method to get the description"""
        if not self.exists(content_index):
            return False
        try:
            params = {"index": content_index, "description": description}
            self.hybridstore.query(
                'MATCH (n:'+self.indexed_label+' {name:$index})'+
                ' SET n.description=$description',
                params = params,
            )
        except Exception:
            return False
        return True

    def get_content_description(self, content_index: str) -> Optional[str]:
        """Method to get the description"""
        if not self.exists(content_index):
            return None
        try:
            params = {"index": content_index}
            result = self.hybridstore.query(
                'MATCH (n:'+self.indexed_label+' {name:$index})'+
                ' RETURN n.description AS description',
                params = params,
            )
            return result.result_set[0][0]
        except Exception:
            return None

    def set_content_metadata(
            self,
            content_index: str,
            metadata: Dict[Any, Any],
        ) -> bool:
        """Method to set the content metadata"""
        if not self.exists(content_index):
            return False
        params = {"index": content_index, "metadata": json.dumps(metadata, indent=2)}
        self.hybridstore.query(
            'MATCH (n:'+self.indexed_label+' {name:$index})'
            +' SET n.metadata=$metadata',
            params = params,
        )
        return True

    def get_content_metadata(self, content_index: str) -> Optional[Dict[Any, Any]]:
        """Method to get the description"""
        if not self.exists(content_index):
            return None
        try:
            params = {"index": content_index}
            result = self.hybridstore.query(
                'MATCH (n:'+self.indexed_label+' {name:$index})'+
                ' RETURN n.metadata AS metadata',
                params = params,
            )
            metadata = result.result_set[0][0]
            return json.loads(metadata)
        except Exception:
            return None
        return None

    def delete_content(self, content_index: str) -> bool:
        """Method to delete an entry"""
        if not self.exists(content_index):
            return False
        params = {"index": content_index}
        self.hybridstore.query(
            'MATCH (n:'+self.indexed_label+' {name:$index}) DELETE n',
            params = params)
        return True
        
    def clear(self):
        """Method to clear the hybridstore"""
        try:
            self.hybridstore.delete()
        except Exception as e:
            pass
        self.init_index()

    def init_index(self):
        try:
            self.hybridstore.query(f"CREATE INDEX FOR (n:`{self.indexed_label}`) ON (n.name)")
            pass
        except Exception as e:
            pass
        try:
            params = {"dim": self.embeddings.dim}
            self.hybridstore.query(
                "CREATE VECTOR INDEX FOR (c:"+self.indexed_label+
                ") ON (c.embeddings_vector) OPTIONS {dimension:$dim, similarityFunction:'euclidean'}",
                params,
            )
        except Exception as e:
            pass


"""The ls command. Copyright (C) 2023 SynaLinks. License: GPL-3.0"""

from typing import List
from .base import BaseShellCommand
from ...hybridstores.filesystem.context import FileSystemContext
from ...hybridstores.filesystem.filesystem import FileSystem
from ...hybridstores.filesystem.path import basename
from ...output_parsers.path import PathOutputParser

class ListDirectory(BaseShellCommand):
    path_parser: PathOutputParser = PathOutputParser()

    def __init__(self, filesystem: FileSystem):
        super().__init__(
            filesystem = filesystem,
            name = "ls",
            description = "list the given directory"
        )

    def execute(self, args: List[str], ctx: FileSystemContext) -> str:
        """Method to list folder/directory"""
        result_list = []
        if len(args)>0:
            path = args[0]
            path = self.path_parser.parse(path)
        else:
            path = ctx.working_directory
        if path.startswith("-"):
            option = args[0]
            raise ValueError(f"Cannot list: Option {option} not supported")
        path = ctx.eval_path(path)
        if self.filesystem.exists(path):
            if not self.filesystem.is_folder(path):
                raise ValueError(f"Cannot list {path}: Not a directory")
        else:
            raise ValueError(f"Cannot list {path}: No such file or directory")
        params = {"path": path}
        result_query = self.filesystem.hybridstore.query(
            'MATCH (f:Folder {name:$path})-[:CONTAINS]->(n:Folder) RETURN n',
            params = params,
        )
        for record in result_query.result_set:
            folder_name = record[0].properties["name"]
            result_list.append(basename(folder_name))
        result_query = self.filesystem.hybridstore.query(
            'MATCH (f:Folder {name:$path})-[:CONTAINS]->(n:Document) RETURN n',
            params = params,
        )
        for record in result_query.result_set:
            document_name = record[0].properties["name"]
            result_list.append(basename(document_name))
        return " ".join(result_list)

    def get_instructions(self) -> str:
        return "The Input should be a unix-like path"


"""The revert trace tool. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import copy
import dspy
from .base import BaseTool
from typing import Optional, Callable
from ..types.state import AgentState
from pydantic import BaseModel

class Integer(BaseModel):
    integer: int

class RevertTraceSignature(dspy.Signature):
    """You will be given an objective, purpose and context
    Using the prompt to help you, you will infer the correct number of steps to revert
    
    Note: Never give an apology or explain what you are doing."""
    objective = dspy.InputField(desc = "The long-term objective (what you are doing)")
    context = dspy.InputField(desc = "The previous actions (what you have done)")
    purpose = dspy.InputField(desc = "The purpose of the action (what you have to do now)")
    prompt = dspy.InputField(desc = "The action specific instructions (How to do it)")
    number: Integer = dspy.OutputField(desc = "The number of steps to revert")

class RevertTraceTool(BaseTool):

    def __init__(
            self,
            agent_state: AgentState,
            lm: Optional[dspy.LM] = None,
        ):
        super().__init__(name = "RevertTrace")
        self.agent_state = agent_state
        self.predict = dspy.TypedPredictor(RevertTraceSignature)

    def revert_trace(self, k: int):
        self.agent_state.program_trace = self.agent_state.program_trace[:len(self.agent_state.program_trace)-k]
        
    def forward(
            self,
            context: str,
            objective: str,
            purpose: str,
            prompt: str,
            disable_inference: bool = False,
        ) -> dspy.Prediction:
        """Method to perform DSPy forward prediction"""
        if not disable_inference:
            pred = self.predict(
                objective = objective,
                context = context,
                purpose = purpose,
                prompt = prompt,
            )
            nb = pred.number.integer
            self.revert_trace(nb)
            return None
        else:
            return None

    def __deepcopy__(self, memo):
        cpy = (type)(self)(
            agent_state = self.agent_state,
            lm = self.lm,
        )
        return cpy



"""The program memory. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import uuid
import os
from falkordb import Node, Graph
from typing import List, Dict, Optional, Tuple
from ..hybridstore import HybridStore, RESERVED_NAMES
from ...embeddings.base import BaseEmbeddings

class ProgramMemory(HybridStore):
    """The program memory"""

    def __init__(
            self,
            index_name: str,
            embeddings: BaseEmbeddings,
            graph_index: str = "program_memory",
            hostname: int = "localhost",
            port: int = 6379,
            username: str = "",
            password: str = "",
            indexed_label: str = "Content",
            wipe_on_start: bool = False,
            ):
        super().__init__(
            index_name = index_name,
            graph_index = graph_index,
            embeddings = embeddings,
            hostname = hostname,
            port = port,
            username = username,
            password = password,
            indexed_label = indexed_label,
            wipe_on_start = wipe_on_start,
        )
        self.playground = self.get_graph("playground")

    def add_texts(
            self,
            texts: List[str],
            ids: List[str] = [],
            descriptions: List[str] = [],
            metadatas: List[Dict[str, str]] = [],
        ):
        """Method to add programs"""
        indexes = []
        dependencies = {}
        for idx, text in enumerate(texts):
            content_index = str(uuid.uuid4().hex) if not ids else ids[idx]
            if descriptions:
                description = descriptions[idx]
            else:
                description = ""
                for line in text.split("\n"):
                    if line.startswith("// @desc:"):
                        description = line.replace("// @desc:", "").strip()
                        break
                    elif line.startswith("//"):
                        pass
                    elif not line:
                        pass
                    else:
                        break
            graph_program = self.get_graph(content_index)
            try:
                graph_program.delete()
            except Exception:
                pass
            try:
                graph_program.query(text)
            except Exception as e:
                raise RuntimeError(f"{content_index}: {e}")
            params = {"index": content_index}
            self.hybridstore.query(
                'MERGE (n:Program {name:$index})',
                params = params,
            )
            metadata = {} if not metadatas else metadatas[idx]
            if description:
                vector = self.embeddings.embed_text(description)
            else:
                vector = self.embeddings.embed_text(text)
            params = {"index": content_index, "vector": vector.tolist()}
            self.hybridstore.query(
               'MERGE (n:'+self.indexed_label+' {name:$index, embeddings_vector:vecf32($vector)})',
               params = params,
            )
            self.set_content(content_index, text)
            if description:
                self.set_content_description(content_index, description)
            if metadata:
                self.set_content_metadata(content_index, metadata)
            params = {"index": content_index}
            self.hybridstore.query(
                "MATCH (p:Program {name:$index}), (c:Content {name:$index}) MERGE (p)-[:CONTAINS]->(c)",
                params = params,
            )
            indexes.append(content_index)
            result = graph_program.query(
                'MATCH (n:Program) RETURN n.program AS program'
            )
            if len(result.result_set) > 0:
                dependencies[content_index] = []
                for record in result.result_set:
                    dependencies[content_index].append(record[0])
        for index, dependency_list in dependencies.items():
            for dependency in dependency_list:
                params = {"index": index, "dependency": dependency}
                self.hybridstore.query(
                    "MATCH (p:Program {name:$index}), (d:Program {name:$dependency}) MERGE (p)-[:DEPENDS_ON]->(d)",
                    params = params,
                )
        return indexes

    def depends_on(self, source: str, target: str):
        """Method to check if a program depend on another"""
        params = {"source": source, "target": target}
        result = self.hybridstore.query(
            "MATCH (n:Program {name:$source})-[r:DEPENDS_ON*]->(m:Program {name:$target}) RETURN r",
            params = params,
        )
        if len(result.result_set) > 0:
            return True
        return False

    def get_program_names(self):
        """Method to get the program names"""
        program_names = []
        result = self.hybridstore.query("MATCH (n:Program) RETURN n.name AS name")
        for record in result.result_set:
            program_names.append(record[0])
        return program_names

    def is_protected(self, program_name: str):
        if program_name in RESERVED_NAMES:
            return True
        else:
            if self.depends_on("main", program_name):
                return True
        return False

    def get_next_node(self, node: Node, program: Graph) -> Optional[Node]:
        """Method to get the next node"""
        try:
            name = node.properties["name"]
        except ValueError:
            raise ValueError("Invalid node: missing required 'name' parameter")
        params = {"name": name}
        result = program.query(
            'MATCH ({name:$name})-[:NEXT]->(m) RETURN m',
            params=params,
        )
        if len(result.result_set) > 0:
            return result.result_set[0][0]
        return None

    def get_starting_node(self, program_name: str) -> Node:
        """Method to get the starting node of the given program"""
        program = self.get_graph(program_name)
        result = program.query(
            'MATCH (n:Control {name:"Start"}) RETURN n')
        if len(result.result_set) == 0:
            raise RuntimeError("No entry point detected,"+
                " please make sure you loaded the programs.")
        if len(result.result_set) > 1:
            raise RuntimeError("Multiple entry point detected,"+
                " please correct your programs.")
        starting_node = result.result_set[0][0]
        return starting_node

    def clear(self):
        """Clear the program memory"""
        programs = self.get_program_names()
        for prog in programs:
            try:
                self.get_graph(prog).delete()
            except:
                pass
        super().clear()


"""The write file tool. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import copy
import dspy
from .base import BaseTool
from typing import Optional
from ..hybridstores.filesystem.filesystem import FileSystem
from ..output_parsers.path import PathOutputParser
from ..output_parsers.prediction import PredictionOutputParser
from ..types.state import AgentState

class WriteFileSignature(dspy.Signature):
    """You will be given an objective, purpose and context
    Using the prompt to help you, you will infer the correct filepath and content
    
    Note: Never give an apology or explain what you are doing."""
    objective = dspy.InputField(desc = "The long-term objective (what you are doing)")
    context = dspy.InputField(desc = "The previous actions (what you have done)")
    purpose = dspy.InputField(desc = "The purpose of the action (what you have to do now)")
    prompt = dspy.InputField(desc = "The action specific instructions (How to do it)")
    filepath = dspy.OutputField(desc = "The path of the file (short and concise)")
    content = dspy.OutputField(desc = "The content to write")

class WriteFileTool(BaseTool):

    def __init__(
            self,
            filesystem: FileSystem,
            agent_state: AgentState,
            lm: Optional[dspy.LM] = None,
        ):
        super().__init__(name = "WriteFile", lm = lm)
        self.predict = dspy.Predict(WriteFileSignature)
        self.agent_state = agent_state
        self.filesystem = filesystem
        self.path_parser = PathOutputParser()
        self.prediction_parser = PredictionOutputParser()

    def write_file(self, filename: str, content: str) -> str:
        try:
            if self.filesystem.is_folder(filename):
                return "Error: Cannot override a directory"
            filename = self.agent_state.context.eval_path(filename)
            metadata = {}
            metadata["filename"] = filename
            self.filesystem.add_texts(
                texts = [content],
                ids = [filename],
                metadatas = [metadata],
            )
            return "Successfully written"
        except Exception as err:
            return str(err)
    
    def forward(
            self,
            context: str,
            objective: str,
            purpose: str,
            prompt: str,
            disable_inference: bool = False,
        ) -> dspy.Prediction:
        """Method to perform DSPy forward prediction"""
        if not disable_inference:
            with dspy.context(lm=self.lm if self.lm is not None else dspy.settings.lm):
                pred = self.predict(
                    objective = objective,
                    context = context,
                    purpose = purpose,
                    prompt = prompt,
                )
            pred.filepath = self.prediction_parser.parse(pred.filepath, prefix="Filename:", stop=["\n"])
            pred.filepath = self.path_parser.parse(pred.filepath)
            pred.content = self.prediction_parser.parse(pred.content, prefix="Content:")
            pred.content = self.prediction_parser.parse(pred.content, prefix="\n\n```\n", stop=["\n```\n\n"])
            dspy.Suggest(
                len(pred.content) != 0,
                "Content must not be empty"
            )
            dspy.Suggest(
                len(pred.filepath) != 0,
                "Filename must not be empty"
            )
            dspy.Suggest(
                len(pred.filepath) < 250,
                "Filename must be short and consice"
            )
            observation = self.write_file(pred.filepath, pred.content)
            return dspy.Prediction(
                filename = pred.filepath,
                content = pred.content,
                observation = observation,
            )
        else:
            raise NotImplementedError("Disabling inference for WriteFile not supported")

    def __deepcopy__(self, memo):
        cpy = (type)(self)(
            filesystem = self.filesystem,
            agent_state = self.agent_state,
            lm = self.lm,
        )
        cpy.predict = copy.deepcopy(self.predict)
        return cpy


from .hybridstores.hybridstore import HybridStore
from .hybridstores.filesystem.context import FileSystemContext
from .hybridstores.filesystem.filesystem import FileSystem
from .hybridstores.program_memory.program_memory import ProgramMemory
from .hybridstores.trace_memory.trace_memory import TraceMemory
from .hybridstores.fact_memory.fact_memory import FactMemory

from .retrievers.document import DocumentRetriever
from .retrievers.program import ProgramRetriever
from .retrievers.action import ActionRetriever
from .retrievers.entity import EntityRetriever

from .text_splitters.base import BaseTextSplitter
from .text_splitters.sentence import SentenceTextSplitter

from .embeddings.base import BaseEmbeddings
from .embeddings.fake import FakeEmbeddings
from .embeddings.sentence_transformer import SentenceTransformerEmbeddings

from .output_parsers.cypher import CypherOutputParser
from .output_parsers.path import PathOutputParser
from .output_parsers.file import FileOutputParser
from .output_parsers.decision import DecisionOutputParser

from .types.actions import (
    AgentAction,
    AgentDecision,
    ProgramCall,
    ProgramEnd,
)
from .types.state import AgentState

from .utility.shell import ShellUtility
from .utility.reader import ReaderUtility
from .utility.archiver import ArchiverUtility
from .utility.tester import TesterUtility
from .utility.browser import BrowserUtility
from .utility.code_interpreter import CodeInterpreterUtility

from .utility.commands import (
    BaseShellCommand,
    ChangeDirectory,
    ListDirectory,
    MakeDirectory,
    Move,
    PrintWorkingDirectory,
    Remove,
    Tree,
)

from .agents.interpreter import GraphProgramInterpreter

from .knowledge_parsers.base import BaseKnowledgeParser
from .knowledge_parsers.python import PythonKnowledgeParser
from .knowledge_parsers.text import TextKnowledgeParser

from .loaders.knowledge import KnowledgeLoader
from .loaders.graph_programs import GraphProgramsLoader

from .hybridagi import HybridAGI

__all__ = [
    CypherOutputParser,
    PathOutputParser,
    FileOutputParser,

    BaseKnowledgeParser,
    PythonKnowledgeParser,
    TextKnowledgeParser,

    KnowledgeLoader,
    GraphProgramsLoader,

    HybridStore,
    FileSystemContext,
    FileSystem,
    ProgramMemory,
    TraceMemory,
    FactMemory,

    DocumentRetriever,
    ProgramRetriever,
    ActionRetriever,
    EntityRetriever,

    GraphProgramInterpreter,

    AgentAction,
    AgentDecision,
    ProgramCall,
    ProgramEnd,
    AgentState,

    BaseEmbeddings,
    FakeEmbeddings,
    SentenceTransformerEmbeddings,

    TesterUtility,
    ArchiverUtility,
    ReaderUtility,
    ShellUtility,
    BrowserUtility,
    CodeInterpreterUtility,

    BaseShellCommand,
    ChangeDirectory,
    ListDirectory,
    MakeDirectory,
    Move,
    PrintWorkingDirectory,
    Remove,
    Tree,

    HybridAGI,
]


"""The trace memory. Copyright (C) 2023 SynaLinks. License: GPL-3.0"""

import uuid
import json
from typing import Union, List
from ...types.actions import AgentAction, AgentDecision, ProgramCall, ProgramEnd
from ...embeddings.base import BaseEmbeddings
from ..hybridstore import HybridStore

class TraceMemory(HybridStore):
    """The trace memory"""

    def __init__(
            self,
            index_name: str,
            embeddings: BaseEmbeddings,
            graph_index: str = "trace_memory",
            hostname: int = "localhost",
            port: int = 6379,
            username: str = "",
            password: str = "",
            indexed_label: str = "Content",
            wipe_on_start: bool = False,
            ):
        super().__init__(
            index_name = index_name,
            graph_index = graph_index,
            embeddings = embeddings,
            hostname = hostname,
            port = port,
            username = username,
            password = password,
            indexed_label = indexed_label,
            wipe_on_start = wipe_on_start,
        )
        self.current_commit = None

    def commit(
            self,
            step: Union[
                AgentAction,
                AgentDecision,
                ProgramCall,
                ProgramEnd,
            ]
        ) -> str:
        new_commit = ""
        content_index = ""
        if isinstance(step, AgentAction):
            new_commit = str(uuid.uuid4().hex)
            # avoid recursive recall by not indexing the past action search
            if step.prediction and step.tool != "PastActionSearch":
                prediction = json.dumps(dict(step.prediction), indent=2)
                indexes = self.add_texts(texts = [prediction])
                content_index = indexes[0]
            else:
                if step.tool == "PastActionSearch":
                    prediction = json.dumps(dict(step.prediction), indent=2)
                else:
                    prediction = "None"
            params = {
                "index" : new_commit,
                "hop" : step.hop,
                "objective" : step.objective,
                "purpose" : step.purpose,
                "tool" : step.tool,
                "prompt" : step.prompt,
                "prediction" : prediction,
                "log" : step.log,
            }
            self.hybridstore.query(
                'MERGE (:Action {name:$index, '+
                'hop:$hop, objective:$objective, '+
                'purpose:$purpose, '+
                'tool:$tool, '+
                'prompt:$prompt, '+
                'prediction:$prediction, '+
                'log:$log})',
                params = params,
            )
            if content_index:
                params = {
                    "commit_index" : new_commit,
                    "content_index": content_index,
                }
                self.hybridstore.query(
                    "MATCH (a:Action {name:$commit_index}), "+
                    "(c:"+self.indexed_label+" {name:$content_index}) MERGE (a)-[:CONTAINS]->(c)",
                    params = params,
                )
        elif isinstance(step, AgentDecision):
            new_commit = str(uuid.uuid4().hex)
            params = {
                "index" : new_commit,
                "hop" : step.hop,
                "objective" : step.objective,
                "purpose" : step.purpose,
                "question" : step.question,
                "options" : step.options,
                "answer" : step.answer,
                "log" : step.log,
            }
            self.hybridstore.query(
                'MERGE (:Decision {name:$index, '+
                'hop:$hop, objective:$objective, '+
                'purpose:$purpose, '+
                'question:$question, '+
                'answer:$answer, '+
                'log:$log})',
                params = params,
            )
        elif isinstance(step, ProgramCall):
            new_commit = str(uuid.uuid4().hex)
            params = {
                "index": new_commit,
                "program": step.program,
                "purpose": step.purpose,
            }
            self.hybridstore.query(
                'MERGE (:ProgramCall {name:$index,'+
                ' program:$program, purpose:$purpose})',
                params = params,
            )
        elif isinstance(step, ProgramEnd):
            new_commit = str(uuid.uuid4().hex)
            params = {"index": new_commit, "program": step.program}
            self.hybridstore.query(
                'MERGE (:ProgramEnd {name:$index, '+
                'program:$program})',
                params = params,
            )
        else:
            raise ValueError("Invalid step provided, could not commit to memory")
        params = {"current": self.current_commit, "new": new_commit}
        if self.current_commit:
            self.hybridstore.query(
                'MATCH (n {name:$current}), '+
                '(m {name:$new}) MERGE (n)-[:NEXT]->(m)',
                params = params,
            )
        self.current_commit = new_commit
        return new_commit if not content_index else content_index

    def start_new_trace(self):
        """Start a new trace"""
        self.current_commit = None

    def get_trace_indexes(
            self,
        ) -> List[str]:
        """Get the traces indexes (the first commit index of each trace)"""
        trace_indexes = []
        result = self.hybridstore.query('MATCH (n:ProgramCall {program:"main"}) RETURN n.name as name')
        for record in result.result_set:
            trace_indexes.append(record[0])
        return trace_indexes

    def is_finished(
            self,
            trace_index: str,
        ) -> bool:
        """Returns True if the execution of the program terminated"""
        params = {"index": trace_index}
        result = self.query(
            'MATCH (n:ProgramCall {name:$index, program:"main"})'
            +'-[:NEXT*]->(m:EndProgram {program:"main"}) RETURN m',
            params = params,
        )
        if len(result.result_set) == 0:
            return False
        return True


import dspy

class HelpfulInteractionSignature(dspy.Signature):
    """Check if the interaction is helpful"""
    context = dspy.InputField(desc="The context for the prediction")
    objective = dspy.InputField(desc="The objective")
    chat_history = dspy.InputField(desc="The chat history")
    helpful = dspy.OutputField(
        desc="Is the interaction helpful for the user?",
        prefix="Helpful[Yes/No]:",
    )

def helpful_interaction(example, prediction, trace=None):
    # This line means that we discard the example if the agent reached the max iterations
    # Meaning it was probably stuck in a loop
    if prediction.finish_reason == "max iters":
        return False
    # Check if the interaction is actually helpful
    pred = dspy.ChainOfThought(HelpfulInteractionSignature)(
        context=prediction.program_trace,
        objective=example.objective,
        chat_history=prediction.chat_history,
    )
    return pred.helpful.lower().strip().strip(".")=="yes"


"""The base text splitter. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import abc
from typing import List

class BaseTextSplitter():

    @abc.abstractmethod
    def split_text(self, text: str) -> List[str]:
        pass



"""The actions. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import dspy
import json
from typing import List, Optional

ACTION_TEMPLATE = \
""" --- Step {hop} ---
Action Purpose: {purpose}
Action: {prediction}"""

DECISION_TEMPLATE = \
""" --- Step {hop} ---
Decision Purpose: {purpose}
Decision Question: {question}
Decision: {answer}"""

CALL_PROGRAM_TEMPLATE = \
""" --- Step {hop} ---
Call Program: {program}
Program Purpose: {purpose}"""

END_PROGRAM_TEMPLATE = \
""" --- Step {hop} ---
End Program: {program}"""

class AgentAction():

    def __init__(
            self,
            hop: int,
            objective: str,
            purpose: str,
            tool: str,
            prompt: str = "",
            prediction: Optional[dspy.Prediction] = None,
            log: str = "",
        ):
        self.hop = hop
        self.objective = objective
        self.purpose = purpose
        self.tool = tool
        self.prompt = prompt
        self.prediction = prediction
        self.log = log

    def __str__(self):
        return ACTION_TEMPLATE.format(
            hop = self.hop,
            purpose = self.purpose,
            prediction = json.dumps(dict(self.prediction), indent=2) if self.prediction else "None",
        )

class AgentDecision():

    def __init__(
            self,
            hop: int,
            objective: str,
            purpose: str,
            question: str,
            options: List[str],
            answer: str,
            prompt: str = "",
            log: str = "",
        ):
        self.hop = hop
        self.objective = objective
        self.purpose = purpose
        self.question = question
        self.options = options
        self.prompt = prompt
        self.answer = answer
        self.log = log

    def __str__(self):
        return DECISION_TEMPLATE.format(
            hop = self.hop,
            purpose = self.purpose,
            question = self.question,
            answer = self.answer,
        )

class ProgramCall():

    def __init__(
            self,
            hop: int,
            purpose: str,
            program: str,
        ):
        self.hop = hop
        self.purpose = purpose
        self.program = program

    def __str__(self):
        return CALL_PROGRAM_TEMPLATE.format(
            hop = self.hop,
            purpose = self.purpose,
            program = self.program,
        )

class ProgramEnd():

    def __init__(
            self,
            hop: int,
            program: str,
        ):
        self.hop = hop
        self.program = program

    def __str__(self):
        return END_PROGRAM_TEMPLATE.format(
            hop = self.hop,
            program = self.program,
        )


"""The base class for shell commands. Copyright (C) 2023 SynaLinks. License: GPL-3.0"""

import abc
from typing import List
from ...hybridstores.filesystem.context import FileSystemContext
from ...hybridstores.filesystem.filesystem import FileSystem

class BaseShellCommand():
    filesystem: FileSystem
    name: str = ""
    description: str = ""

    def __init__(
            self,
            name: str,
            description: str,
            filesystem: FileSystem,
        ):
        self.name = name
        self.description = description
        self.filesystem = filesystem

    @abc.abstractmethod
    def execute(self, args: List[str], ctx: FileSystemContext) -> str:
        pass

    @abc.abstractmethod
    def get_instructions(self) -> str:
        pass

    def run(self, args: List[str], ctx: FileSystemContext)-> str:
        """Run the command"""
        if len(args) > 0:
            if args[0] == "--help":
                return self.get_instructions()
        return self.execute(args, ctx)


"""The agent state. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

from collections import deque, defaultdict
from falkordb import Node, Graph
from typing import List, Iterable, Optional, Tuple
from hybridagi import FileSystemContext

class AgentState():

    def __init__(self):
        self.init()

    def get_current_program(self) -> Optional[Graph]:
        """Method to retreive the current program from the stack"""
        if len(self.program_stack) > 0:
            return self.program_stack[-1][1]
        return None

    def get_current_node(self) -> Optional[Node]:
        """Method to retreive the current node from the stack"""
        if len(self.program_stack) > 0:
            return self.program_stack[-1][0]
        return None
    
    def set_current_node(self, node: Node):
        """Method to set the current node from the stack"""
        if len(self.program_stack) > 0:
            _, program = self.program_stack[-1]
            self.program_stack[-1] = (node, program)

    def call_program(self, starting_node: Node, program: Graph):
        """Append the program stack"""
        self.program_stack.append((starting_node, program))

    def end_program(self) -> Optional[Graph]:
        """Pop the program stack"""
        _, program = self.program_stack.pop()
        return program

    def update_variable(self, key:str, value:str):
        """Update a variable"""
        self.variables[key] = value

    def init(self):
        self.objective = "N/A"
        self.current_hop = 0
        self.program_trace = []
        self.program_stack = deque()
        self.context = FileSystemContext()
        self.chat_history = []
        self.variables = {}
        self.plots = []


"""The Cypher output parser. Copyright (C) 2023 SynaLinks. License: GPL-3.0"""

from .base import BaseOutputParser

class CypherOutputParser(BaseOutputParser):
    """
    The output parser for RedisGraph Cypher queries
    """

    def parse(self, output:str) -> str:
        """Fix and validate the given Cypher query."""
        output = output.strip()
        output = output.strip("\"'`")
        output = output.rstrip(".,;")
        return output


"""The program name output parser. Copyright (C) 2023 SynaLinks. License: GPL-3.0"""

import re
from .base import BaseOutputParser

class ProgramNameOutputParser(BaseOutputParser):
    """
    The Output Parser for program names
    """
    def parse(self, output:str) -> str:
        """Fix and validate the given Cypher query."""
        output = output.replace("\"", "")
        output = output.replace("\\_", "_")
        output = output.strip("\"'. \n")
        output = output.replace(".cypher", "")
        output = re.sub('(?!^)([A-Z]+)', r'_\1', output).lower()
        return output.strip()


from typing import Optional
import tree_sitter_python as tspython
from tree_sitter import Language, Parser, Tree, Node
from .base import BaseKnowledgeParser
from ..hybridstores.fact_memory.fact_memory import FactMemory
from ..hybridstores.filesystem.filesystem import FileSystem

PY_LANGUAGE = Language(tspython.language())

class PythonKnowledgeParser(BaseKnowledgeParser):

    def __init__(
            self,
            filesystem: FileSystem,
            fact_memory: FactMemory,
        ):
        super().__init__(
            filesystem = filesystem,
            fact_memory = fact_memory,
            valid_extensions = [".py"],
        )
        self.parser = Parser(PY_LANGUAGE)
        if filesystem is None:
            raise ValueError("")

    def parse(self, dest_path:str, content: str):
        self.filesystem.add_texts(
            texts=[content],
            ids=[dest_path],
        )
        src = content.encode('utf-8')
        tree = self.parser.parse(src)
        # Get class names and body
        query = PY_LANGUAGE.query("""
(class_definition
    name: (identifier) @class-name
    body: (block) @class-body)
"""
        )
        matches = query.matches(tree.root_node)
        for record in matches:
            class_name = src[record[1]["class-name"].start_byte:record[1]["class-name"].end_byte]
            class_body = src[record[1]["class-body"].start_byte:record[1]["class-body"].end_byte]
            class_name = class_name.decode('utf8')
            class_body = class_body.decode('utf8')
            metadata = {}
            metadata["filepath"] = dest_path
            self.fact_memory.add_texts(
                texts = [class_body],
                ids = [class_name],
                descriptions = [class_name],
                metadatas = [metadata],
            )
        query = PY_LANGUAGE.query("""
(class_definition
    name: (identifier) @class-name
    superclasses: (argument_list
        (identifier) @super-class))
"""
        )
        matches = query.matches(tree.root_node)
        for record in matches:
            class_name = src[record[1]["class-name"].start_byte:record[1]["class-name"].end_byte]
            super_class = src[record[1]["super-class"].start_byte:record[1]["super-class"].end_byte]
            class_name = class_name.decode('utf8')
            super_class = super_class.decode('utf8')
            self.fact_memory.add_triplet(class_name, "inherits from", super_class)
        query = PY_LANGUAGE.query("""
(class_definition
name: (identifier) @class-name
body: (block
    (function_definition
    name: (identifier) @method-name
    body: (block) @method-body)))
"""
        )
        matches = query.matches(tree.root_node)
        node_class = ""
        method_name = ""
        for record in matches:
            class_name = src[record[1]["class-name"].start_byte:record[1]["class-name"].end_byte]
            method_name = src[record[1]["method-name"].start_byte:record[1]["method-name"].end_byte]
            method_body = src[record[1]["method-body"].start_byte:record[1]["method-body"].end_byte]
            class_name = class_name.decode('utf8')
            method_name = method_name.decode('utf8')
            method_body = method_body.decode('utf8')
            metadata = {}
            metadata["filepath"] = filepath
            self.fact_memory.add_texts(
                texts = [method_body],
                ids = [class_name+":"+method_name],
                descriptions = [class_name+":"+method_name],
                metadatas = [metadata],
            )
            self.fact_memory.add_triplet(class_name, "has method", class_name+":"+method_name)

    def read(self, source_path: str) -> str:
        f = open(source_path, "r")
        file_content = f.read()
        return file_content


"""The shell utility. Copyright (C) 2023 SynaLinks. License: GPL-3.0"""

from typing import List, Dict
from ..hybridstores.filesystem.filesystem import FileSystem
from .commands.base import BaseShellCommand
from ..types.state import AgentState

class ShellUtility():
    """The internal shell for the filesystem"""

    def __init__(
            self,
            filesystem: FileSystem,
            agent_state: AgentState,
            commands: List[BaseShellCommand]):
        """The shell utility constructor"""
        self.commands_map = {cmd.name:cmd for cmd in commands}
        self.filesystem = filesystem
        self.agent_state = agent_state

    def execute(self, args:List[str]) -> str:
        """Execute a shell-like command"""
        if len(args) > 0:
            cmd = args[0]
            arguments = args[1:]
            if cmd in self.commands_map.keys():
                return self.commands_map[cmd].run(arguments, self.agent_state.context)
            else:
                raise ValueError(f"Command '{cmd}' not supported")
        else:
            raise ValueError(
                "Please use one of the following commands: "+
                str(list(self.commands_map.keys())))




import abc
import dspy
import copy
from typing import Optional, Union, Callable, Dict, Any
from dspy.signatures.signature import ensure_signature

class BaseTool(dspy.Module):

    def __init__(
            self,
            name: str,
            lm: Optional[dspy.LM] = None,
        ):
        self.name = name
        self.lm = lm

    @abc.abstractmethod
    def forward(
            self,
            trace: str,
            objective: str,
            purpose: str,
            prompt: str,
            disable_inference: bool = False,
        ) -> dspy.Prediction:
        pass

class Tool(BaseTool):

    def __init__(
            self,
            name: str,
            signature: Union[str, dspy.Signature],
            func: Callable[..., Dict[Any, Any]],
            instructions: str = "",
            lm: Optional[dspy.LM] = None,
        ):
        super().__init__(name=name, lm=lm)
        self.signature = signature = ensure_signature(signature)
        self.instructions = instructions

        self.input_fields = self.signature.input_fields
        self.output_fields = self.signature.output_fields

        signature_dict = {}

        signature_dict["objective"] = dspy.InputField(
            prefix = "Objective:",
            desc = "The long-term objective (what you are doing)",
        )
        signature_dict["context"] = dspy.InputField(
            prefix = "Context:",
            desc = "The previous actions (what you have done)",
        )
        signature_dict["purpose"] = dspy.InputField(
            prefix = "Purpose:",
            desc = "The purpose of the action (what you have to do now)",
        )
        signature_dict["prompt"] = dspy.InputField(
            prefix = "Prompt:",
            desc = "The action specific instructions (How to do it)",
        )

        for k in self.input_fields.keys():
            signature_dict[k] = dspy.OutputField(
                prefix = k.title()+":",
            )

        outputs = ", ".join([f"`{k}`" for k in self.input_fields.keys()])

        instr = []
        if instructions:
            instr = [instructions]

        instr.extend([
            "You will be given an objective, purpose and context",
            f"Using the prompt to help you, you will infer the correct {outputs}"
        ])

        instr = "\n".join(instr)
        self.predict = dspy.Predict(dspy.Signature(signature_dict, instr))
        self.func = func

    def forward(
            self,
            context: str,
            objective: str,
            purpose: str,
            prompt: str,
            disable_inference: bool = False,
        ):
        if not disable_inference:
            with dspy.context(lm=self.lm if self.lm is not None else dspy.settings.lm):
                pred = self.predict(
                    objective = objective,
                    context = context,
                    purpose = purpose,
                    prompt = prompt,
                )
            args = {}
            for k in self.input_fields.keys():
                args[k] = pred[k]
            result = self.func(**args)
            args.update(result)
            return dspy.Prediction(
                **args
            )
        else:
            raise NotImplementedError("Tool does not support disabling inferences")

    def __deepcopy__(self, memo):
        cpy = (type)(self)(
            name = self.name,
            signature = self.signature,
            func = self.func,
            instructions = self.instructions,
            lm = self.lm,
        )
        cpy.predict = copy.deepcopy(self.predict)
        return cpy




"""The tree command. Copyright (C) 2023 SynaLinks. License: GPL-3.0"""

from typing import List, Tuple
from .base import BaseShellCommand
from ...hybridstores.filesystem.context import FileSystemContext
from ...hybridstores.filesystem.filesystem import FileSystem
from ...hybridstores.filesystem.path import basename
from ...output_parsers.path import PathOutputParser

ELBOW = "└──"
TEE = "├──"
PIPE_PREFIX = "│   "
SPACE_PREFIX = "    "

class Tree(BaseShellCommand):

    def __init__(self, filesystem: FileSystem):
        super().__init__(
            filesystem = filesystem,
            name = "tree",
            description = "display the given directory structure",
        )
        self.path_parser = PathOutputParser()

    def execute(self, args: List[str], ctx: FileSystemContext) -> str:
        """Method to display the directory structure"""
        result_list = []
        if len(args)>0:
            path = args[0]
            path = self.path_parser.parse(path)
        else:
            path = ctx.working_directory
        if path.startswith("-"):
            option = args[0]
            raise ValueError(f"Cannot use tree: Option {option} not supported")
        path = ctx.eval_path(path)
        if self.filesystem.exists(path):
            if not self.filesystem.is_folder(path):
                raise ValueError(f"Cannot use tree {path}: Not a directory")
        else:
            raise ValueError(f"Cannot use tree {path}: No such file or directory")
        
        result_list.append(path)
        result_list, folder_count, file_count = self._get_tree(path, result_list, "", 0, 0)
        result_list.append(f"{folder_count} directories, {file_count} files")
        return "\n".join(result_list)

    def get_instructions(self) -> str:
        return "The Input should be a unix-like path"

    def _get_tree(
            self: str,
            path: str,
            result: List[str],
            prefix: str,
            folder_count: int,
            file_count: int,
        ) -> Tuple[List[str], int, int]:
        """Recursively gather the tree structure"""
        files = self._get_files(path)
        directories = self._get_subdirectories(path)

        for i, item in enumerate(directories):
            is_last = i == len(directories) - 1
            prefix_str = ELBOW if is_last else TEE
            result.append(f"{prefix}{prefix_str} {basename(item)}")
            folder_count += 1
            new_prefix = prefix + SPACE_PREFIX if is_last else prefix + PIPE_PREFIX
            result, folder_count, file_count = self._get_tree(item, result, new_prefix, folder_count, file_count)

        for i, item in enumerate(files):
            is_last = i == len(files) - 1
            prefix_str = ELBOW if is_last else TEE
            result.append(f"{prefix}{prefix_str} {basename(item)}")
            file_count += 1
        
        return result, folder_count, file_count

    def _get_subdirectories(self, path: str) -> List[str]:
        """Retrieve the folders of the given folder."""
        params = {"path": path}
        query = 'MATCH (f:Folder {name:$path})-[:CONTAINS]->(n:Folder) RETURN n'
        result_query = self.filesystem.hybridstore.query(query, params=params)
        return sorted(record[0].properties["name"] for record in result_query.result_set)

    def _get_files(self, path: str) -> List[str]:
        """Retrieve the files of the given folder."""
        params = {"path": path}
        query = 'MATCH (f:Folder {name:$path})-[:CONTAINS]->(n:Document) RETURN n'
        result_query = self.filesystem.hybridstore.query(query, params=params)
        return sorted(record[0].properties["name"] for record in result_query.result_set)


"""The sentence embeddings. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""
# Modified from dspy vectorizer

import numpy as np
from typing import Union, List
from .base import BaseEmbeddings

class SentenceTransformerEmbeddings(BaseEmbeddings):

    def __init__(
            self,
            model_name_or_path: str,
            dim: int,
            max_gpu_devices: int = 1,
            batch_size: int = 256,
            max_seq_length: int = 256,
            normalize_embeddings: bool = True,
        ):
        super().__init__(dim=dim)
        self.model_name_or_path = model_name_or_path
        self.max_gpu_devices = max_gpu_devices
        self.batch_size = batch_size
        self.max_seq_length= max_seq_length
        self.normalize_embeddings = normalize_embeddings
        try:
            from sentence_transformers import SentenceTransformer
        except ImportError:
            raise ImportError(
                "You need to install sentence_transformers library to use pretrained embeddings. "
                "Please run `pip install sentence-transformers`",
            )
        try:
            from dsp.utils.ann_utils import determine_devices
        except ImportError:
            raise ImportError(
                "You need to install faiss library to use pretrained embeddings. "
                "Please run `pip install faiss-gpu` or `pip install faiss-cpu`",
            )
        self.num_devices, self.is_gpu = determine_devices(self.max_gpu_devices)
        self.proxy_device = 'cuda' if self.is_gpu else 'cpu'

        self.model = SentenceTransformer(self.model_name_or_path, device=self.proxy_device)
        self.model.max_seq_length = self.max_seq_length

    def embed_text(self, query_or_queries: Union[str, List[str]]) -> np._typing.NDArray:
        text_to_vectorize = [query_or_queries] if not isinstance(query_or_queries, list) else query_or_queries

        if self.is_gpu and self.num_devices > 1:
            target_devices = list(range(self.num_devices))
            pool = self.model.start_multi_process_pool(target_devices=target_devices)
            # Compute the embeddings using the multi-process pool
            emb = self.model.encode_multi_process(
                sentences=text_to_vectorize,
                pool=pool,
                batch_size=self.batch_size,
            )
            self.model.stop_multi_process_pool(pool)
            # for some reason, multi-GPU setup doesn't accept normalize_embeddings parameter
            if self.normalize_embeddings:
                emb = emb / np.linalg.norm(emb)
        else:
            emb = self.model.encode(
                sentences=text_to_vectorize,
                batch_size=self.batch_size,
                normalize_embeddings=self.normalize_embeddings,
            )
        if not isinstance(query_or_queries, list):
            return emb[0]
        else:
            return emb
    
    def embed_image(self, image_or_images: Union[np._typing.NDArray, List[np._typing.NDArray]]) -> np._typing.NDArray:
        raise NotImplementedError("SentenceTransformer does not support image embeddings")


"""The ask user tool. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import copy
from typing import List, Optional, Callable, Optional
import json
import dspy
from .base import BaseTool
from ..types.state import AgentState
from ..output_parsers.prediction import PredictionOutputParser

class AskUserSignature(dspy.Signature):
    """You will be given an objective, purpose and context
    Using the prompt to help you, you will infer the correct question"""
    objective = dspy.InputField(desc = "The long-term objective (what you are doing)")
    context = dspy.InputField(desc = "The previous actions (what you have done)")
    purpose = dspy.InputField(desc = "The purpose of the action (what you have to do now)")
    prompt = dspy.InputField(desc = "The action specific instructions (how to do it)")
    question = dspy.OutputField(desc = "The question to ask to the user")

class SimulateAnswerSignature(dspy.Signature):
    """Answer from the perspective of a real person, if you don't known imagine what a user with the given profile would answer"""
    objective = dspy.InputField(desc = "The long-term objective (what you are doing)")
    user_profile = dspy.InputField(desc = "The user profile")
    chat_history = dspy.InputField(desc = "The chat history")
    question = dspy.InputField(desc = "The question to assess")
    user_answer = dspy.OutputField(desc = "The answer from the perspective of a real person (only few words)")

class AskUserTool(BaseTool):

    def __init__(
            self,
            agent_state: AgentState,
            user_profile: str = "An average user",
            ask_user_func: Optional[Callable[[str], str]] = None,
            num_history: int = 50,
            simulated: bool = True,
            lm: Optional[dspy.LM] = None,
        ):
        super().__init__(name = "AskUser", lm = lm)
        self.predict = dspy.Predict(AskUserSignature)
        self.simulated = simulated
        self.simulate = dspy.Predict(SimulateAnswerSignature)
        self.agent_state = agent_state
        self.num_history = num_history
        self.ask_user_func = ask_user_func
        self.prediction_parser = PredictionOutputParser()

    def ask_user(self, question : str) -> str:
        if self.ask_user_func:
            return self.ask_user_func(question)
        else:
            raise ValueError(
                "You should specify a function to call to use `AskUser` outside simulation")

    def simulate_ask_user(self, question: str):
        chat_history = json.dumps(self.agent_state.chat_history[:-self.num_history], indent=2)
        pred = self.simulate(
            objective = self.agent_state.objective,
            chat_history = chat_history,
            question = question,
        )
        with dspy.context(lm=self.lm if self.lm is not None else dspy.settings.lm):
            pred.answer = self.prediction_parser.parse(
                pred.user_answer,
                prefix = "User Answer:",
            )
        return pred.answer

    def forward(
            self,
            context: str,
            objective: str,
            purpose: str,
            prompt: str,
            disable_inference: bool = False,
        ) -> dspy.Prediction:
        """Method to perform DSPy forward prediction"""
        if not disable_inference:
            with dspy.context(lm=self.lm if self.lm is not None else dspy.settings.lm):
                pred = self.predict(
                    objective = objective,
                    context = context,
                    purpose = purpose,
                    prompt = prompt,
                )
            pred.question = self.prediction_parser.parse(
                pred.question,
                prefix = "Question:",
            )
            self.agent_state.chat_history.append(
                {"role": "AI", "message": pred.question}
            )
            if self.simulated:
                answer = self.simulate_ask_user(pred.question)
            else:
                answer = self.ask_user(pred.question)
            self.agent_state.chat_history.append(
                {"role": "User", "message": answer}
            )
            return dspy.Prediction(
                question = pred.question,
                answer = answer,
            )
        else:
            self.agent_state.chat_history.append(
                {"role": "AI", "message": prompt}
            )
            if self.simulated:
                answer = self.simulate_ask_user(prompt)
            else:
                answer = self.ask_user(message)
            self.agent_state.chat_history.append(
                {"role": "User", "message": answer}
            )
            return dspy.Prediction(
                question = prompt,
                answer = answer,
            )

    def __deepcopy__(self, memo):
        cpy = (type)(self)(
            agent_state = self.agent_state,
            num_history = self.num_history,
            ask_user_func = self.ask_user_func,
            simulated = self.simulated,
            lm = self.lm,
        )
        cpy.predict = copy.deepcopy(self.predict)
        return cpy


"""The document retriever. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import dspy
from typing import Union, Optional, List
from dsp.utils import dotdict
from ..embeddings.base import BaseEmbeddings
from ..hybridstores.filesystem.filesystem import FileSystem

class DocumentRetriever(dspy.Retrieve):
    """Retrieve document chunks based on similarity"""

    def __init__(
            self,
            filesystem: FileSystem,
            embeddings: BaseEmbeddings,
            distance_threshold: float = 1.2,
            k: int = 3,
        ):
        """The retriever constructor"""
        super().__init__(k = k)
        self.filesystem = filesystem
        self.embeddings = embeddings
        self.distance_threshold = distance_threshold

    def forward(
            self,
            query_or_queries: Union[str, List[str]],
            k:Optional[int] = None,
        ) -> dspy.Prediction:
        """Method to perform DSPy forward prediction"""
        if not isinstance(query_or_queries, list):
            query_or_queries = [query_or_queries]
        query_vectors = self.embeddings.embed_text(query_or_queries)
        contents = []
        indexes = {}
        for vector in query_vectors:
            # For an obscure reason falkordb needs a bigger k to find more indexed items
            params = {"indexed_label": self.filesystem.indexed_label, "vector": list(vector), "k": 2*int(k or self.k)}
            query = " ".join([
                'CALL db.idx.vector.queryNodes($indexed_label, "embeddings_vector", $k, vecf32($vector)) YIELD node, score',
                'RETURN node.name AS name, score'])
            result = self.filesystem.hybridstore.query(
                query,
                params = params,
            )
            if len(result.result_set) > 0:
                for record in result.result_set:
                    if record[0] not in indexes:
                        indexes[record[0]] = True
                    else:
                        continue
                    content = self.filesystem.get_content(record[0])
                    metadata = self.filesystem.get_content_metadata(record[0])
                    distance = float(record[1])
                    if distance < self.distance_threshold:
                        contents.extend([{"passage": dotdict({"long_text": content, "metadata": metadata}), "distance": distance}])
        sorted_passages = sorted(
            contents,
            key=lambda x: x["distance"],
            reverse=False,
        )[: k or self.k]
        return dspy.Prediction(
            passages=[el["passage"] for el in sorted_passages]
        )


import re
from .base import BaseOutputParser
from typing import List

class PredictionOutputParser(BaseOutputParser):
    """
    The output parser for predictions
    """

    def parse(self, output: str, prefix: str, stop: List[str] = []) -> str:
        first_occurence = output.find(prefix)
        if first_occurence >= 0:
            output = output[first_occurence+len(prefix):]
        for s in stop:
            first_occurence = output.find(s)
            if first_occurence >= 0:
                output = output[:first_occurence]
        output = output.strip()
        return output



"""The mkdir command. Copyright (C) 2023 SynaLinks. License: GPL-3.0"""

from typing import List
from .base import BaseShellCommand
from ...hybridstores.filesystem.context import FileSystemContext
from ...hybridstores.filesystem.filesystem import FileSystem
from ...hybridstores.filesystem.path import dirname
from ...output_parsers.path import PathOutputParser

class MakeDirectory(BaseShellCommand):
    path_parser: PathOutputParser = PathOutputParser()

    def __init__(self, filesystem: FileSystem):
        super().__init__(
            filesystem = filesystem,
            name = "mkdir",
            description = "make a new directory",
        )

    def execute(self, args: List[str], ctx: FileSystemContext) -> str:
        """Method to create a directory"""
        if len(args)==0:
            raise ValueError(
                "Cannot create directory: Missing operand."+
                " Try `mkdir --help` for more information."
                )
        nb_dir = 0
        args = args[0].split()
        for path in args:
            path = self.path_parser.parse(path)
            if path.startswith("-"):
                raise ValueError("Cannot create directory: Options not supported")
            path = ctx.eval_path(path)
            if self.filesystem.exists(path):
                raise ValueError(f"Cannot create directory {path}: File exists")
            parent_folder = dirname(path)
            if not self.filesystem.exists(parent_folder):
                raise ValueError(
                    f"Cannot create directory {path}:"+
                    f" '{parent_folder}' No such file or directory"
                )
            self.filesystem.create_folder(path)
            nb_dir += 1
        if nb_dir == 1:
            return f"Sucessfully created directory {path}"
        return f"Sucessfully created {nb_dir} directories"
 
    def get_instructions(self) -> str:
        return "The Input should be a unix-like path of the folder to be created"


"""The program tester. Copyright (C) 2023 SynaLinks. License: GPL-3.0"""

from typing import List
from ..hybridstores.program_memory.program_memory import ProgramMemory

FORBIDDEN_COMMANDS = [
    "DETACH",
    "DELETE",
    "SET",
    "WHEN",
    "WITH",
    "AS",
    "RETURN",
]

class TesterUtility():
    """The class to make some verification to graph programs"""

    def __init__(
            self,
            program_memory: ProgramMemory,
    ):
        self.program_memory = program_memory
        self.program_name = ""

    def check_injection_attemp(self, program:str):
        for cmd in FORBIDDEN_COMMANDS:
            if cmd in program:
                raise RuntimeError("Error: Detected Cypher injection attemp")

    def load_program_into_playground(
            self,
            program_name: str,
            program: str):
        """Try to load a program into playground for testing"""
        try:
            self.program_memory.playground.delete()
        except Exception:
            pass
        try:
            self.program_memory.playground.query(program)
        except Exception as err:
            raise RuntimeError(
                f"Error occured: {err}. "+\
                "Invalid Cypher query, please correct it")
        self.program_name = program_name

    def check_protected_program(self, program_name: str):
        """Check if the system try to modify a protected program"""
        if self.program_memory.is_protected(program_name):
            raise RuntimeError(
                "Error occured: "+\
                "Trying to modify a protected program. "+\
                "Change the name of the program")

    def check_starting_node_count(self):
        """Check how many starting node the program have"""
        result = self.program_memory.playground.query(
            'MATCH (n:Control {name:"Start"}) RETURN n')
        if len(result.result_set) == 0:
            raise RuntimeError(
                "Error occured: "+\
                "No starting node detected, please "+\
                "make sure to start your program correctly")
        if len(result.result_set) > 1:
            raise RuntimeError(
                "Error occured: "+\
                "Starting node is connected to more than one node "+\
                "make sure to start your program correctly")

    def check_ending_node_count(self):
        """Check how many ending node the program have"""
        result = self.program_memory.playground.query(
            'MATCH (n:Control {name:"End"}) RETURN n')
        if len(result.result_set) == 0:
            raise RuntimeError(
                "Error occured: "+\
                "No ending node detected, please "+\
                "make sure to end your program correctly")
        if len(result.result_set) > 1:
            raise RuntimeError(
                "Error occured: "+\
                "Multiple ending point detected, "+
                "please correct your programs.")

    def check_starting_node_output(self):
        """Check starting node outgoing edges"""
        result = self.program_memory.playground.query(
            'MATCH (n:Control {name:"Start"})-[r]->(m) RETURN r')
        if len(result.result_set) == 0:
            raise RuntimeError(
                "Error occured: "+\
                "Starting node is not connected (meaning it doesn't have outgoing edge) "+\
                "make sure to start your program correctly")
        if len(result.result_set) > 1:
            raise RuntimeError(
                "Error occured: "+\
                "Starting node is connected to more than one node "+\
                "make sure to start your program correctly")

    def check_starting_node_input(self):
        """Check starting node incoming edges"""
        result = self.program_memory.playground.query(
            'MATCH (m)-[r]->(n:Control {name:"Start"}) RETURN r')
        if len(result.result_set) > 0:
            raise RuntimeError(
                "Error occured: "+\
                "Starting node cannot have incoming edges "+\
                "make sure to start your program correctly")

    def check_ending_node_input(self):
        """Check ending node incoming edges"""
        result = self.program_memory.playground.query(
            'MATCH (m)-[r]->(n:Control {name:"End"}) RETURN r')
        if len(result.result_set) == 0:
            raise RuntimeError(
                "Error occured: "+\
                "Ending node should at least have one incoming edge"+\
                "make sure to end your program correctly")

    def check_ending_node_output(self):
        """Check ending node outgoing edges"""
        result = self.program_memory.playground.query(
            'MATCH (n:Control {name:"End"})-[r]->(m) RETURN r')
        if len(result.result_set) > 0:
            raise RuntimeError(
                "Error occured: "+\
                "Ending node cannot have outgoing edges"+\
                "make sure to end your program correctly")

    def check_program_dependencies(self):
        result = self.program_memory.playground.query(
            'MATCH (p:Program) RETURN p.program AS program')
        for record in result.result_set:
            subprogram = record[0]
            if self.program_memory.is_protected(subprogram):
                raise RuntimeError(
                    f"Error while loading '{self.program_name}'. "+\
                    f"Trying to call a protected program '{subprogram}'. Try to remove it")
            if not self.program_memory.exists(subprogram):
                raise RuntimeError(
                    "Error occured: "+\
                    f"The sub-program '{subprogram}' does not exist. "+\
                    "Please correct your program")
                
    def verify_programs(
            self,
            names: List[str],
            programs: List[str]):
        """Verify programs"""
        for idx, program in enumerate(programs):
            program_name = names[idx]
            self.check_protected_program(program_name)
            self.check_injection_attemp(program)

            self.load_program_into_playground(
                program_name,
                program,
            )
            
            self.check_starting_node_count()
            self.check_starting_node_input()
            self.check_starting_node_output()

            self.check_ending_node_count()
            self.check_ending_node_input()
            self.check_ending_node_output()

            self.check_program_dependencies()


"""The update objective tool. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import copy
import dspy
from .base import BaseTool
from typing import Optional, Callable
from ..types.state import AgentState

class ClearTraceTool(BaseTool):

    def __init__(
            self,
            agent_state: AgentState,
        ):
        super().__init__(name = "ClearTrace")
        self.agent_state = agent_state
        
    def forward(
            self,
            context: str,
            objective: str,
            purpose: str,
            prompt: str,
            disable_inference: bool = False,
        ) -> dspy.Prediction:
        """Method to perform DSPy forward prediction"""
        self.agent_state.program_trace = []
        return None

    def __deepcopy__(self, memo):
        cpy = (type)(self)(
            agent_state = self.agent_state,
        )
        return cpy



"""The duckduckgo tool. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import copy
import dspy
from .base import BaseTool
from typing import Optional
from duckduckgo_search import DDGS
from ..output_parsers.query import QueryOutputParser
from ..output_parsers.prediction import PredictionOutputParser

class BrowseWebsiteSignature(dspy.Signature):
    """You will be given an objective, purpose and context
    Using the prompt to help you, you will infer the correct url"""
    objective = dspy.InputField(desc = "The long-term objective (what you are doing)")
    context = dspy.InputField(desc = "The previous actions (what you have done)")
    purpose = dspy.InputField(desc = "The purpose of the action (what you have to do now)")
    prompt = dspy.InputField(desc = "The action specific instructions (How to do it)")
    url = dspy.OutputField(desc = "The website to browse")

class BrowseWebsiteTool(BaseTool):

    def __init__(
            self,
            lm: Optional[dspy.LM] = None,
        ):
        super().__init__(name = "BrowseWebsite", lm = lm)
        self.predict = dspy.Predict(BrowseWebsiteSignature)
        self.prediction_parser = PredictionOutputParser()
    
    def forward(
            self,
            context: str,
            objective: str,
            purpose: str,
            prompt: str,
            disable_inference: bool = False,
            k: Optional[int] = None,
        ) -> dspy.Prediction:
        """Method to perform DSPy forward prediction"""
        if not disable_inference:
            with dspy.context(lm=self.lm if self.lm is not None else dspy.settings.lm):
                pred = self.predict(
                    objective = objective,
                    context = context,
                    purpose = purpose,
                    prompt = prompt,
                )
            pred.url = self.prediction_parser.parse(pred.url, prefix="Url:", stop=["\n"])
            result = self.browser.browse_website(pred.url)
            return dspy.Prediction(
                url = pred.url,
                result = result,
            )
        else:
            result = self.browser.browse_website(prompt)
            return dspy.Prediction(
                url = prompt,
                result = result,
            )

    def __deepcopy__(self, memo):
        cpy = (type)(self)(
            lm = self.lm,
        )
        cpy.predict = copy.deepcopy(self.predict)
        return cpy


"""The fake embeddings. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import numpy as np
from typing import Union, List
from .base import BaseEmbeddings

class FakeEmbeddings(BaseEmbeddings):
    
    def __init__(
            self,
            dim: int,
            normalize_embeddings: bool = True,
        ):
        super().__init__(dim=dim)
        self.normalize_embeddings = normalize_embeddings

    def embed_text(self, query_or_queries: Union[str, List[str]]) -> np._typing.NDArray:
        if not isinstance(query_or_queries, list):
            emb = np.random.random(self.dim)
            if self.normalize_embeddings:
                return emb / np.linalg.norm(emb)
            else:
                return emb
        else:
            embeddings = []
            for _ in query_or_queries:
                emb = np.random.random(self.dim)
                if self.normalize_embeddings:
                    embeddings.append(emb / np.linalg.norm(emb))
                else:
                    embeddings.append(emb)
            return embeddings
    
    def embed_image(self, image_or_images: Union[np._typing.NDArray, List[np._typing.NDArray]]) -> np._typing.NDArray:
        if not isinstance(image_or_images, list):
            emb = np.random.random(self.dim)
            if self.normalize_embeddings:
                return emb / np.linalg.norm(emb)
            else:
                return emb
        else:
            embeddings = []
            for _ in image_or_images:
                emb = np.random.random(self.dim)
                if self.normalize_embeddings:
                    embeddings.append(emb / np.linalg.norm(emb))
                else:
                    embeddings.append(emb)
            return embeddings


"""The cd command. Copyright (C) 2023 SynaLinks. License: GPL-3.0"""

from typing import List
from .base import BaseShellCommand
from ...hybridstores.filesystem.context import FileSystemContext
from ...hybridstores.filesystem.filesystem import FileSystem
from ...output_parsers.path import PathOutputParser

class ChangeDirectory(BaseShellCommand):
    path_parser: PathOutputParser = PathOutputParser()

    def __init__(
            self,
            filesystem: FileSystem,
        ):
        super().__init__(
            filesystem = filesystem,
            name = "cd",
            description = "change the current working directory",
        )

    def execute(self, args: List[str], ctx: FileSystemContext) -> str:
        """Method to change directory"""
        if len(args)>0:
            path = args[0]
            path = self.path_parser.parse(path)
            path = ctx.eval_path(path)
            if self.filesystem.exists(path):
                if self.filesystem.is_file(path):
                    raise ValueError(
                        f"Cannot change directory {path}: Not a directory")
            else:
                raise ValueError(
                    f"Cannot change directory {path}: No such file or directory")
            ctx.previous_working_directory = ctx.working_directory
            ctx.working_directory = path
            return f"Successfully changed working directory {path}"
        raise ValueError("Cannot change directory: Input path not provided.")

    def get_instructions(self) -> str:
        return "The Input should be a unix-like path."


"""The read file tool. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import copy
import dspy
from typing import Optional
from .base import BaseTool
from ..types.state import AgentState
from ..utility.code_interpreter import CodeInterpreterUtility
from ..output_parsers.prediction import PredictionOutputParser

class CodeInterpreterSignature(dspy.Signature):
    """You will be given an objective, purpose and context
    Using the prompt to help you, you will infer the correct code
    
    Note: Never ask for user input and ensure that your last make sure the last line in your code evaluates to the correct value"""
    objective = dspy.InputField(desc = "The long-term objective (what you are doing)")
    context = dspy.InputField(desc = "The previous actions (what you have done)")
    purpose = dspy.InputField(desc = "The purpose of the action (what you have to do now)")
    prompt = dspy.InputField(desc = "The action specific instructions (How to do it)")
    code = dspy.OutputField(desc = "The code to write")

class CodeInterpreterTool(BaseTool):

    def __init__(
            self,
            code_interpreter: Optional[CodeInterpreterUtility] = None,
            lm: Optional[dspy.LM] = None,
        ):
        super().__init__(name = "CodeInterpreter", lm = lm)
        self.predict = dspy.Predict(CodeInterpreterSignature)
        if code_interpreter:
            self.code_interpreter = code_interpreter
        else:
            self.code_interpreter = CodeInterpreterUtility()
        self.prediction_parser = PredictionOutputParser()
    
    def forward(
            self,
            context: str,
            objective: str,
            purpose: str,
            prompt: str,
            disable_inference: bool = False,
        ) -> dspy.Prediction:
        """Method to perform DSPy forward prediction"""
        if not disable_inference:
            with dspy.context(lm=self.lm if self.lm is not None else dspy.settings.lm):
                pred = self.predict(
                    objective = objective,
                    context = context,
                    purpose = purpose,
                    prompt = prompt,
                )
            pred.code = pred.code.replace("\\n", "\n")
            pred.code = self.prediction_parser.parse(pred.code,
                prefix="\n\n```python\n",
                stop = ["\n```", "\n```\n\n"],
            )
            pred.code = self.prediction_parser.parse(pred.code,
                prefix="```python\n",
                stop = ["\n```", "\n```\n\n"],
            )
            code, output, err = self.code_interpreter.execute_code(pred.code)
            if not err:
                result = output
            else:
                result = err
            return dspy.Prediction(
                code = pred.code,
                observation = result,
            )
        else:
            code, output, err = self.code_interpreter.execute_code(prompt)
            if not err:
                result = output
            else:
                result = err
            return dspy.Prediction(
                code = prompt,
                observation = result,
            )

    def __deepcopy__(self, memo):
        cpy = (type)(self)(
            code_interpreter = self.code_interpreter,
            lm = self.lm,
        )
        cpy.predict = copy.deepcopy(self.predict)
        return cpy
        


import os
from typing import List
from ..hybridstores.filesystem.context import FileSystemContext
from ..hybridstores.fact_memory.fact_memory import FactMemory
from ..hybridstores.filesystem.filesystem import FileSystem
from ..knowledge_parsers.base import BaseKnowledgeParser

class KnowledgeLoader():

    def __init__(
            self,
            filesystem: FileSystem,
            fact_memory: FactMemory,
            parsers: List[BaseKnowledgeParser] = []
        ):
        self.filesystem = filesystem
        self.fact_memory = fact_memory
        self.valid_extensions = []
        self.parsers = {}
        for parser in parsers:
            self.valid_extensions.extend(parser.valid_extensions)
            for extension in parser.valid_extensions:
                self.parsers[extension] = parser

    def from_folders(
            self,
            folders: List[str],
        ):
        for i, folder in enumerate(folders):
            folder_name = os.path.basename(folder)
            ctxt = FileSystemContext()
            folder_name = ctxt.eval_path(folder_name)
            self.filesystem.create_folder(folder_name)
            for dirpath, dirnames, filenames in os.walk(folder):
                if "venv" in dirpath or dirpath.startswith("__") > 0 or dirpath.startswith(".") > 0:
                    continue
                for filename in filenames:
                    if not filename.startswith("."):
                        _, extension = os.path.splitext(filename)
                        if extension in self.valid_extensions:
                            parser = self.parsers[extension]
                            source_path = os.path.join(dirpath, filename)
                            target_path = os.path.join(dirpath.replace(folder, folder_name), filename)
                            if extension in self.valid_extensions:
                                try:
                                    file_content = parser.read(source_path)
                                    parser.parse(target_path, file_content)
                                except Exception:
                                    continue


"""The reader utility. Copyright (C) 2023 SynaLinks. License: GPL-3.0"""

from ..hybridstores.filesystem.filesystem import FileSystem

class ReaderUtility():
    """The reader utility"""

    def __init__(
            self,
            filesystem: FileSystem):
        """The reader constructor"""
        self.filesystem = filesystem
        self.current_consulted_document = ""
        self.last_content_consulted = ""

    def read_document(self, path:str) -> str:
        """
        Method to read a document.
        This method display only one content at a time.
        Use it multiple times with the same target to scroll.
        """
        if self.filesystem.exists(path):
            if self.filesystem.is_folder(path):
                return f"Cannot read {path}: Is a directory."
        else:
            return f"Cannot read {path}: No such file or directory"
        if self.current_consulted_document != "":
            if self.current_consulted_document == path:
                content_key = self.filesystem.get_next(self.last_content_consulted)
                next_key = self.filesystem.get_next(content_key)
            else:
                content_key = self.filesystem.get_beginning_of_file(path)
                next_key = self.filesystem.get_next(content_key)
        else:
            content_key = self.filesystem.get_beginning_of_file(path)
            next_key = self.filesystem.get_next(content_key)
        
        if next_key != "":
            self.current_consulted_document = path
        else:
            self.current_consulted_document = ""
        content = str(self.filesystem.get_content(content_key))
        metadata = self.filesystem.get_content_metadata(content_key)
        self.last_content_consulted = content_key
        if next_key != "":
            return content + "\n\n[...]" + (f"\n\n{metadata}" if metadata else "")
        else:
            return content + (f"\n\n{metadata}" if metadata else "")


"""The filesystem memory. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import os
from ..hybridstore import HybridStore
from typing import List, Dict, Optional, Any
from .path import dirname, join

from ...text_splitters.sentence import SentenceTextSplitter

from .context import FileSystemContext
from ...embeddings.base import BaseEmbeddings

class FileSystem(HybridStore):
    """The filesystem memory"""

    def __init__(
        self,
        index_name: str,
        embeddings: BaseEmbeddings,
        context: Optional[FileSystemContext] = None,
        graph_index: str = "filesystem",
        hostname: str = "localhost",
        port: int = 6379,
        username: str = "",
        password: str = "",
        indexed_label: str = "Content",
        wipe_on_start: bool = False,
        chunk_size: int = 1024,
        chunk_overlap: int = 0,
    ):
        super().__init__(
            index_name = index_name,
            graph_index = graph_index,
            embeddings = embeddings,
            hostname = hostname,
            port = port,
            username = username,
            password = password,
            indexed_label = indexed_label,
            wipe_on_start = wipe_on_start, 
        )
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.init()

    def add_texts(
            self,
            texts: List[str],
            ids: List[str] = [],
            descriptions: List[str] = [],
            metadatas: List[Dict[Any, Any]] = [],
        ) -> List[str]:
        """Method to add texts"""
        indexes = []
        assert(len(texts) == len(ids))
        if metadatas:
            assert(len(texts) == len(metadatas))
        for idx, text in enumerate(texts):
            path = ids[idx]
            metadata = metadatas[idx] if metadatas else {}
            metadata["filename"] = path
            if self.exists(path):
                self.remove_documents([path])
            text_splitter = \
            SentenceTextSplitter(
                chunk_size=self.chunk_size,
                chunk_overlap=self.chunk_overlap,
            )
            subdocs_texts = text_splitter.split_text(text)
            if metadata:
                subdocs_metadatas = [metadata for _ in range(len(subdocs_texts))]
            else:
                subdocs_metadatas = []
            indexes = super().add_texts(
                texts = subdocs_texts,
                metadatas = subdocs_metadatas,
            )
            if len(indexes) > 0:
                self.create_document(path)
                for i, key in enumerate(indexes):
                    params = {"path": path, "index": key}
                    self.hybridstore.query(
                        'MATCH (n:Document {name:$path}),'+
                        ' (m:Content {name:$index}) MERGE (n)-[:CONTAINS]->(m)',
                        params = params,
                    )
                    if i > 0:
                        params = {"previous_index": indexes[i-1], "index": key}
                        self.hybridstore.query(
                            'MATCH (n:Content {name:$previous_index}),'+
                            ' (m:Content {name:$index}) MERGE (n)-[:NEXT]->(m)',
                            params = params,
                        )
                params = {"path": path, "index": indexes[0]}
                self.hybridstore.query(
                    'MATCH (n:Document {name:$path}),'+
                    ' (m:Content {name:$index}) MERGE (n)-[:BOF]->(m)',
                    params = params,
                )
                params = {"path": path, "last_index": indexes[0]}
                self.hybridstore.query(
                    'MATCH (n:Document {name:$path}),'+
                    ' (m:Content {name:$last_index}) MERGE (n)-[:EOF]->(m)',
                    params = params,
                )
        return indexes

    def append_texts(
            self,
            texts: List[str],
            ids: List[str] = [],
            descriptions: List[str] = [],
            metadatas: List[Dict[Any, Any]] = [],
        ) -> List[str]:
        """Method to append texts"""
        indexes = []
        assert(len(texts) == len(ids))
        for idx, text in enumerate(texts):
            path = ids[idx]
            metadata = metadatas[idx] if metadatas else {}
            if not self.exists(path):
                self.add_texts(
                    texts = [text],
                    ids = [path],
                    metadatas = [metadata],
                )
            else:
                text_splitter = \
                SentenceTextSplitter(
                    chunk_size=self.chunk_size,
                    chunk_overlap=self.chunk_overlap,
                )
                subdocs_texts = text_splitter.split_text(text)
                if metadata:
                    subdocs_metadatas = [metadata for _ in range(subdocs_texts)]
                else:
                    subdocs_metadatas = []
                indexes = super().add_texts(
                    texts = subdocs_texts,
                    metadatas = subdocs_metadatas,
                )
                eof = self.get_end_of_file(path)
                if len(indexes) > 0:
                    for i, index in enumerate(indexes):
                        params = {"path": path, "index": index}
                        self.hybridstore.query(
                            'MATCH (n:Document {name:$path}),'+
                            ' (m:Content {name:$index}) MERGE (n)-[:CONTAINS]->(m)',
                            params = params,
                        )
                        if i > 0:
                            params = {"previous_index": indexes[i-1], "index": index}
                            self.hybridstore.query(
                                'MATCH (n:Content {name:$previous_index}),'+
                                ' (m:Content {name:$index}) MERGE (n)-[:NEXT]->(m)',
                                params = params,
                            )
                    params = {"eof_index": eof, "first_index": indexes[0]}
                    self.hybridstore.query(
                        'MATCH (n:Content {name:$eof_index}),'+
                        ' (m:Content {name:$first_index}) MERGE (n)-[:NEXT]->(m)',
                        params = params,
                    )
                    params = {"path": path}
                    self.hybridstore.query(
                        'MATCH (n:Document {name:$path})-[r:EOF]->() DELETE r',
                        params = params,
                    )
                    params = {"path": path, "last_index": indexes[-1]}
                    self.hybridstore.query(
                        'MATCH (n:Document {name:$path}),'+
                        ' (m:Content {name:$last_index}) MERGE (n)-[:EOF]->(m)',
                        params = params,
                    )
        return indexes

    def get_document(self, path:str) -> str:
        """
        Method to get the full document.
        Do not use it for the LLM tools as it can exceed the max token.
        """
        if self.exists(path, label = "Document"):
            if self.is_folder(path):
                return "Cannot read directory."
        else:
            return "No such file or directory."
        text = ""
        bof = self.get_beginning_of_file(path)
        eof = self.get_end_of_file(path)
        if bof == eof:
            params = {"bof_index": bof}
            result = self.hybridstore.query(
                'MATCH (n:Content {name:$bof_index}) RETURN n',
                params = params,
            )
            content_key = result.result_set[0][0].properties["name"]
            content = self.get_content(content_key)
            text = content
        else:
            params = {"bof_index": bof, "eof_index": eof}
            result = self.hybridstore.query(
                'MATCH path=(:Content {name:$bof_index})'+
                '-[:NEXT*]->(:Content {name:$eof_index}) RETURN nodes(path)',
                params = params,
            )
            all_content = []
            for node in result.result_set[0][0]:
                content_key = node.properties["name"]
                content = self.get_content(content_key)
                all_content.append(content)
            text = "\n".join(all_content)
        return text

    def create_document(self, path:str):
        """Method to create a document (no check performed)"""
        params = {"path": path}
        self.hybridstore.query('MERGE (:Document {name:$path})', params = params)
        parent_folder = dirname(path)
        while not parent_folder:
            params = {"parent_folder": parent_folder, "path": path}
            self.hybridstore.query(
                'MATCH (n:Folder {name:$parent_folder}),'+
                ' (m:Document {name:$path}) MERGE (n)-[:CONTAINS]->(m)',
                params = params,
            )
            parent_folder = dirname(path)
        params = {"parent_folder": parent_folder, "path": path}
        self.hybridstore.query(
            'MATCH (n:Folder {name:$parent_folder}),'+
            ' (m:Document {name:$path}) MERGE (n)-[:CONTAINS]->(m)',
            params = params,
        )

    def remove_documents(self, paths: List[str]):
        for path in paths:
            indexes = []
            params = {"path": path}
            result = self.hybridstore.query(
                'MATCH (c:Content)<-[:CONTAINS]-(d:Document {name:$path}) RETURN c.name AS name',
                params = params,
            )
            for record in result.result_set:
                indexes.append(record[0])
            params = {"path": path}
            self.hybridstore.query(
                'MATCH (c:Content)<-[:CONTAINS]-(d:Document {name:$path}) DELETE c, d',
                params = params,
            )

    def create_folder(self, path:str):
        """Method to create a folder (no check performed)"""
        params = {"path": path}
        self.hybridstore.query('MERGE (n:Folder {name:$path})', params = params)
        parent_folder = dirname(path)
        params = {"parent_folder": parent_folder, "path": path}
        self.hybridstore.query(
            'MATCH (n:Folder {name:$parent_folder}),'+
            ' (m:Folder {name:$path}) MERGE (n)-[:CONTAINS]->(m)',
            params = params,
        )

    def is_folder(self, path:str) -> bool:
        """Method to check if a folder at path is present in the hybridstore"""
        params = {"path": path}
        result = self.hybridstore.query('MATCH (n:Folder {name:$path}) RETURN n',
            params = params,
        )
        if len(result.result_set) > 0:
            return True
        return False

    def is_empty(self, path:str) -> bool:
        """Method to check if a folder at path is empty in the hybridstore"""
        params = {"path": path}
        result = self.hybridstore.query(
            'MATCH (n:Folder {name:$path})-[:CONTAINS]->(:Document) RETURN n',
            params = params,
        )
        if len(result.result_set) == 0:
            return True
        return False

    def is_file(self, path:str) -> bool:
        """Method to check if a file at path is present in the hybridstore"""
        params = {"path": path}
        result = self.hybridstore.query('MATCH (n:Document {name:$path}) RETURN n',
            params = params,
        )
        if len(result.result_set) > 0:
            return True
        return False

    def is_beginning_of_file(self, content_index:str) -> bool:
        """Returns True if the content is the beginning of the file"""
        params = {"index": content_index}
        result = self.hybridstore.query(
            'MATCH (n:Content {name:$index})<-[:BOF]-() RETURN n',
            params = params,
        )
        if len(result.result_set) > 0:
            return True
        return False

    def is_end_of_file(self, content_index:str) -> bool:
        """Returns True if the content is the end of the file"""
        params = {"index": content_index}
        result = self.hybridstore.query(
            'MATCH (n:Content {name:$index})<-[:EOF]-() RETURN n')
        if len(result.result_set) > 0:
            return True
        return False
    
    def get_beginning_of_file(self, path: str) -> str:
        """Method to return the BOF content index"""
        params = {"path": path}
        result = self.hybridstore.query(
            'MATCH (:Document {name:$path})-[:BOF]->(n:Content) RETURN n',
            params = params,
        )
        return result.result_set[0][0].properties["name"]

    def get_end_of_file(self, path: str) -> str:
        """Method to return the EOF content index"""
        params = {"path": path}
        result = self.hybridstore.query(
            'MATCH (:Document {name:$path})-[:EOF]->(n:Content) RETURN n',
            params = params,
        )
        return result.result_set[0][0].properties["name"]

    def get_next(self, content_index: str) -> str:
        """Method to return the next content index"""
        params = {"index": content_index}
        result = self.hybridstore.query(
            'MATCH (:Content {name:index})-[:NEXT]->(n:Content) RETURN n',
            params = params,
        )
        if len(result.result_set) > 0:
            return result.result_set[0][0].properties["name"]
        return ""

    def init(self):
        """Method to initialize the filesystem"""
        self.hybridstore.query('MERGE (:Folder {name:"/"})')
        self.create_folder("/home")
        self.create_folder("/home/user")
        self.create_folder("/home/user/Downloads")
        self.create_folder("/home/user/Documents")
        self.create_folder("/home/user/Pictures")
        self.create_folder("/home/user/Music")

    def clear(self):
        super().clear()
        self.init()


"""The internal shell tool. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import shlex
import copy
import dspy
from .base import BaseTool
from typing import Optional
from ..hybridstores.filesystem.filesystem import FileSystem
from ..utility.shell import ShellUtility
from ..output_parsers.prediction import PredictionOutputParser
from ..types.state import AgentState
from ..utility.commands import (
    ChangeDirectory,
    ListDirectory,
    MakeDirectory,
    Move,
    PrintWorkingDirectory,
    Remove,
    Tree,
)

class InternalShellSignature(dspy.Signature):
    """You will be given an objective, purpose and context
    Using the prompt to help you, you will infer the correct unix shell command, whitelist: [`cd`, `ls`, `mkdir`, `mv`, `pwd`, `rm`, `tree`]
    
    Note: Never apologies or explain what to do, only give the command"""
    objective = dspy.InputField(desc = "The long-term objective (what you are doing)")
    context = dspy.InputField(desc = "The previous actions (what you have done)")
    purpose = dspy.InputField(desc = "The purpose of the action (what you have to do now)")
    prompt = dspy.InputField(desc = "The action specific instructions (How to do it)")
    unix_shell_command = dspy.OutputField(desc = "The unix command between [`cd`, `ls`, `mkdir`, `mv`, `pwd`, `rm`, `tree`]")

class InternalShellTool(BaseTool):

    def __init__(
            self,
            filesystem: FileSystem,
            agent_state: AgentState,
            lm: Optional[dspy.LM] = None,
        ):
        super().__init__(name = "InternalShell")
        self.predict = dspy.Predict(InternalShellSignature)
        self.agent_state = agent_state
        self.filesystem = filesystem
        self.shell = ShellUtility(
            filesystem = self.filesystem,
            agent_state = self.agent_state,
            commands = [
                ChangeDirectory(filesystem=filesystem),
                ListDirectory(filesystem=filesystem),
                MakeDirectory(filesystem=filesystem),
                Move(filesystem=filesystem),
                PrintWorkingDirectory(filesystem=filesystem),
                Remove(filesystem=filesystem),
                Tree(filesystem=filesystem),
            ]
        )
        self.prediction_parser = PredictionOutputParser()

    def execute(self, command: str) -> str:
        command = command.strip("`")
        s = shlex.shlex(command, punctuation_chars=True)
        args = list(s)
        invalid_symbols = ["|", "||", "&", "&&", ">", ">>", "<", "<<", ";"]
        invalid = False
        if len(list(set(invalid_symbols).intersection(args))) > 0:
            invalid = True
            dspy.Suggest( invalid,
                "Piping, redirection and multiple commands are not supported:"+
                " Use one command at a time, without semicolon."
            )
        try:
            return self.shell.execute(args)
        except Exception as err:
            return str(err)
    
    def forward(
            self,
            context: str,
            objective: str,
            purpose: str,
            prompt: str,
            disable_inference: bool = False,
        ) -> dspy.Prediction:
        """Method to perform DSPy forward prediction"""
        if not disable_inference:
            with dspy.context(lm=self.lm if self.lm is not None else dspy.settings.lm):
                pred = self.predict(
                    objective = objective,
                    context = context,
                    purpose = purpose,
                    prompt = prompt,
                )
            pred.unix_shell_command = self.prediction_parser.parse(
                pred.unix_shell_command,
                prefix = "Unix Shell Command:",
                stop = ["\n"],
            )
            pred.unix_shell_command = pred.unix_shell_command.strip("`")
            observation = self.execute(pred.unix_shell_command)
            return dspy.Prediction(
                unix_shell_command = pred.unix_shell_command,
                observation = observation,
            )
        else:
            observation = self.execute(prompt)
            return dspy.Prediction(
                unix_shell_command = prompt,
                observation = observation,
            )

    def __deepcopy__(self, memo):
        cpy = (type)(self)(
            filesystem = self.filesystem,
            agent_state = self.agent_state,
            lm = self.lm,
        )
        cpy.predict = copy.deepcopy(self.predict)
        return cpy


"""The speak tool. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import copy
from typing import List, Optional, Callable
from colorama import Fore, Style
import dspy
from .base import BaseTool
from ..types.state import AgentState
from ..output_parsers.prediction import PredictionOutputParser

class SpeakSignature(dspy.Signature):
    """You will be given an objective, purpose and context
    Using the prompt to help you, you will infer the correct message to send to the user
    
    Note: Never give an apology or explain what you are doing."""
    objective = dspy.InputField(desc = "The long-term objective (what you are doing)")
    context = dspy.InputField(desc = "The previous actions (what you have done)")
    purpose = dspy.InputField(desc = "The purpose of the action (what you have to do now)")
    prompt = dspy.InputField(desc = "The action specific instructions (How to do it)")
    message = dspy.OutputField(desc = "The message to send to the user")

class SpeakTool(BaseTool):

    def __init__(
            self,
            agent_state: AgentState,
            speak_func: Optional[Callable[[str], None]] = None,
            simulated: bool = True,
            lm: Optional[dspy.LM] = None,
        ):
        super().__init__(name = "Speak", lm = lm)
        self.predict = dspy.Predict(SpeakSignature)
        self.simulated = simulated
        self.agent_state = agent_state
        self.speak_func = speak_func
        self.prediction_parser = PredictionOutputParser()

    def speak(self, message: str):
        if self.speak_func:
            return self.speak_func(message)
        else:
            raise ValueError(
                "You should specify a function to call to use `Speak` outside simulation"
            )

    def forward(
            self,
            context: str,
            objective: str,
            purpose: str,
            prompt: str,
            disable_inference: bool = False,
        ) -> dspy.Prediction:
        """Method to perform DSPy forward prediction"""
        if not disable_inference:
            with dspy.context(lm=self.lm if self.lm is not None else dspy.settings.lm):
                pred = self.predict(
                    objective = objective,
                    context = context,
                    purpose = purpose,
                    prompt = prompt,
                )
            pred.message = self.prediction_parser.parse(
                pred.message,
                prefix = "Message:",
            )
            self.agent_state.chat_history.append(
                {"role": "AI", "message": pred.message}
            )
            if not self.simulated:
                self.speak(pred.message)
            return dspy.Prediction(
                message = pred.message
            )
        else:
            self.agent_state.chat_history.append(
                {"role": "AI", "message": prompt}
            )
            if not self.simulated:
                self.speak(prompt)
            return dspy.Prediction(
                message = prompt
            )

    def __deepcopy__(self, memo):
        cpy = (type)(self)(
            agent_state = self.agent_state,
            speak_func = self.speak_func,
            simulated = self.simulated,
            lm = self.lm,
        )
        cpy.predict = copy.deepcopy(self.predict)
        return cpy


from .base import BaseShellCommand
from .cd import ChangeDirectory
from .ls import ListDirectory
from .mkdir import MakeDirectory
from .mv import Move
from .pwd import PrintWorkingDirectory
from .rm import Remove
from .tree import Tree

__all__ = [
    "BaseShellCommand",
    "ChangeDirectory",
    "ListDirectory",
    "MakeDirectory",
    "Move",
    "PrintWorkingDirectory",
    "Remove",
    "Tree",
]


"""The text splitter functions. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""
# modifed from llama-index
# The MIT License

# Copyright (c) Jerry Liu

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import re
from typing import Callable, List
from .base import BaseTextSplitter

def truncate_text(text: str, text_splitter: BaseTextSplitter) -> str:
    """Truncate text to fit within the chunk size."""
    chunks = text_splitter.split_text(text)
    return chunks[0]


def split_text_keep_separator(text: str, separator: str) -> List[str]:
    """Split text with separator and keep the separator at the end of each split."""
    parts = text.split(separator)
    result = [separator + s if i > 0 else s for i, s in enumerate(parts)]
    return [s for s in result if s]


def split_by_sep(sep: str, keep_sep: bool = True) -> Callable[[str], List[str]]:
    """Split text by separator."""
    if keep_sep:
        return lambda text: split_text_keep_separator(text, sep)
    else:
        return lambda text: text.split(sep)


def split_by_char() -> Callable[[str], List[str]]:
    """Split text by character."""
    return lambda text: list(text)


def split_by_regex(regex: str) -> Callable[[str], List[str]]:
    """Split text by regex."""
    return lambda text: re.findall(regex, text)


def split_by_phrase_regex() -> Callable[[str], List[str]]:
    """Split text by phrase regex.

    This regular expression will split the sentences into phrases,
    where each phrase is a sequence of one or more non-comma,
    non-period, and non-semicolon characters, followed by an optional comma,
    period, or semicolon. The regular expression will also capture the
    delimiters themselves as separate items in the list of phrases.
    """
    regex = "[^,.;。]+[,.;。]?"
    return split_by_regex(regex)



import dspy

class FactualAnswerSignature(dspy.Signature):
    """Check if the answer is present in the provided trace"""
    context = dspy.InputField(desc="The context for the prediction")
    question = dspy.InputField(desc="Question to be answered")
    answer = dspy.InputField(desc="Answer for the question")
    correct = dspy.OutputField(
        desc="Is the answer factually correct based on the context?",
        prefix="Correct[Yes/No]:",
    )

def factual_answer(example, prediction, trace=None):
    # This line means that we discard the example if the agent reached the max iterations
    # Meaning it was probably stuck in a loop
    if prediction.finish_reason == "max iters":
        return False
    # Check if the answer is actually based on the context
    pred = dspy.ChainOfThought(FactualAnswerSignature)(
        context = prediction.program_trace,
        question = example.objective,
        answer = prediction.final_answer,
    )
    return pred.correct.lower().strip().strip(".")=="yes"


"""The path functions. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

from typing import List

def join(paths: List[str]) -> str:
    return '/'.join(paths)

def dirname(path:str) -> str:
    if path == "/":
        return path
    else:
        tokens = path.split('/')
        dirname = '/'.join(tokens[:-1])
        if dirname == "":
            return "/"
        return dirname

def basename(path:str) -> str:
    tokens = path.split('/')
    return tokens[-1]


import re
import builtins
import io
from .python_interpreter import CodePrompt, PythonInterpreter

DEFAULT_WHITELIST = [
    "numpy",
    "math",
]

class PrintWrapper(io.StringIO):
  def __call__(self, *args, **kwargs):
    # Pass the object instance (self) as the file
    return builtins.print(*args, file=self, **kwargs)

class CodeInterpreterUtility():

    def __init__(self, import_white_list=DEFAULT_WHITELIST):
        self.import_white_list = import_white_list

    def parse_code(self, code: str):
        code_match = re.search(r"```python[ \n](.*?)[ \n]```?", code, re.DOTALL)
        code_block = (code_match.group(1) if code_match else code).replace("\\n", "\n")
        if not code_block:
            return code, "Error: Empty code after parsing."
        if "\n" not in code_block and code_block.count("=") > 1:
            return code, "Error: Code format is not correct."
        lines = code_block.split("\n")
        last_line_match = re.match(r"^(\w+)\s*=", lines[-1].strip())
        if last_line_match and len(lines) > 1:
            code_block += "\n" + last_line_match.group(1)
        else:
            code_block = re.sub(
                r"([a-zA-Z_]\w* *=.*?)(?=[a-zA-Z_]\w* *=)", r"\1\n", code_block,
            )
            code_block = re.sub(
                r"([a-zA-Z_]\w* *=.*?)([a-zA-Z_]\w*)$", r"\1\n\2", code_block,
            )
        return code_block, None

    def execute_code(self, code: str):
        parsed_code, error = self.parse_code(code)
        if error:
            return code, None, error
        if not code:
            return code, None, "Error: Empty code before execution."
        code_prompt = CodePrompt(code, code_type="python")
        # print = PrintWrapper()
        interpreter = PythonInterpreter(action_space={}, import_white_list=self.import_white_list)
        try:
            output = str(code_prompt.execute(interpreter=interpreter)[0])
            # msg = print.getvalue()
            return code, output, None
        except Exception as e:
            return code, None, str(e)


from .critique_answer import critique_answer
from .critique_trace import critique_trace
from .factual_answer import factual_answer
from .helpful_interaction import helpful_interaction

__all__ = [
    critique_answer,
    critique_trace,
    factual_answer,
    helpful_interaction,
]


"""The list query output parser. Copyright (C) 2023 SynaLinks. License: GPL-3.0"""

from .base import BaseOutputParser
from typing import List, Union

class ListQueryOutputParser(BaseOutputParser):
    """
    The Output Parser for search queries
    """
    def parse(self, output:str) -> Union[str, List[str]]:
        """Fix the given query"""
        output_list = output.split(" or ")
        if len(output_list) > 1:
            return [o.replace("\"", "").strip() for o  in output_list]
        else:
            return output.replace("\"", "").strip()


import os
import abc
from typing import List, Optional
from ..hybridstores.filesystem.filesystem import FileSystem
from ..hybridstores.fact_memory.fact_memory import FactMemory

class BaseKnowledgeParser():

    def __init__(
            self,
            filesystem: FileSystem,
            fact_memory: FactMemory,
            valid_extensions: List[str] = [],
        ):
        self.filesystem = filesystem
        self.fact_memory = fact_memory
        self.valid_extensions = valid_extensions

    @abc.abstractmethod
    def parse(self, dest_path: str, content: str):
        pass

    def read(self, source_path: str) -> str:
        pass



"""The duckduckgo tool. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import dspy
import copy
from .base import BaseTool
from typing import Optional
from duckduckgo_search import DDGS
from ..output_parsers.query import QueryOutputParser
from ..output_parsers.prediction import PredictionOutputParser

class DuckDuckGoSearchSignature(dspy.Signature):
    """You will be given an objective, purpose and context
    Using the prompt to help you, you will infer the correct DuckDuckGo query"""
    objective = dspy.InputField(desc = "The long-term objective (what you are doing)")
    context = dspy.InputField(desc = "The previous actions (what you have done)")
    purpose = dspy.InputField(desc = "The purpose of the action (what you have to do now)")
    prompt = dspy.InputField(desc = "The action specific instructions (How to do it)")
    query = dspy.OutputField(desc = "The DuckDuckGo search query (only few words)")

class DuckDuckGoSearchTool(BaseTool):

    def __init__(
            self,
            k: int = 5,
            lm: Optional[dspy.LM] = None,
        ):
        super().__init__(name = "DuckDuckGoSearch", lm = lm)
        self.predict = dspy.Predict(DuckDuckGoSearchSignature)
        self.k = k
        self.query_parser = QueryOutputParser()
        self.prediction_parser = PredictionOutputParser()
    
    def forward(
            self,
            context: str,
            objective: str,
            purpose: str,
            prompt: str,
            disable_inference: bool = False,
            k: Optional[int] = None,
        ) -> dspy.Prediction:
        """Method to perform DSPy forward prediction"""
        if not disable_inference:
            with dspy.context(lm=self.lm if self.lm is not None else dspy.settings.lm):
                pred = self.predict(
                    objective = objective,
                    context = context,
                    purpose = purpose,
                    prompt = prompt,
                )
            pred.query = self.prediction_parser.parse(pred.query, prefix="Query:", stop=["\n"])
            pred.query = self.query_parser.parse(pred.query)
            result = DDGS().text(pred.query, max_results=k if k else self.k)
            return dspy.Prediction(
                search_query = pred.query,
                results = result,
            )
        else:
            result = DDGS().text(prompt, max_results=k if k else self.k)
            return dspy.Prediction(
                search_query = prompt,
                results = result,
            )

    def __deepcopy__(self, memo):
        cpy = (type)(self)(
            k = self.k,
            lm = self.lm,
        )
        cpy.predict = copy.deepcopy(self.predict)
        return cpy


"""The base embeddings. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import abc
import numpy as np
from typing import Union, List

class BaseEmbeddings():
    
    def __init__(self, dim: int):
        self.dim = dim

    @abc.abstractmethod
    def embed_text(self, query_or_queries: Union[str, List[str]]) -> np._typing.NDArray:
        pass
    
    @abc.abstractmethod
    def embed_image(self, image_or_images: Union[np._typing.NDArray, List[np._typing.NDArray]]) -> np._typing.NDArray:
        pass


"""The browser utility. Copyright (C) 2023 SynaLinks. License: GPL-3.0"""

import re
from typing import Tuple, Dict, Any, List
import requests
from ..hybridstores.filesystem.filesystem import FileSystem

class BrowserUtility():

    def __init__(
            self,
            filesystem: FileSystem,
            chunk_size: int = 1024,
            chunk_overlap: int = 0,
        ):
        self.filesystem = filesystem
        self.current_consulted_url = ""
        self.last_url_consulted = ""
        self.current_chunks = []
        self.current_chunk_index = 0
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap

    def scrap_text(
            self,
            url: str,
        ) -> str:
        """Method to scrap the text from an url using jina.ai"""
        response = requests.get("https://r.jina.ai/"+url)
        return response.text

    def browse_website(
            self,
            url: str,
        ) -> str:
        """
        Method to browse a website url.
        This method display only one chunk at a time.
        Use it multiple times with the same target to scroll.
        """
        scrap = False
        if self.current_consulted_url != "":
            if self.current_consulted_url != url:
                scrap = True
        else:
            scrap = True
        if scrap:
            page_content = self.scrap_text(url)
            text_splitter = \
            SentenceTextSplitter(
                chunk_size=self.chunk_size,
                chunk_overlap=self.chunk_overlap,
            )
            chunks = text_splitter.split_text(text)
            self.current_chunks = chunks
            self.current_chunk_index = 0
        else:
            self.current_chunk_index = (self.current_chunk_index + 1) % len(self.current_chunks)
        self.current_consulted_url = url

        content = self.current_chunks[self.current_chunk_index].page_content
        metadata = self.current_chunks[self.current_chunk_index].metadata

        result_string = content
        if self.current_chunk_index != len(self.current_chunks)-1:
            result_string + "\n\n[...]"
        else:
            result_string
        return result_string


"""The call program tool. Copyright (C) 2024 SynaLinks. License: GPL-3.0"""

import copy
import dspy
from .base import BaseTool
from typing import Optional, Callable
from ..hybridstores.program_memory.program_memory import ProgramMemory
from ..types.state import AgentState
from ..output_parsers.program_name import ProgramNameOutputParser
from ..output_parsers.prediction import PredictionOutputParser

class CallProgramSignature(dspy.Signature):
    """You will be given an objective, purpose and context
    Using the prompt to help you, you will infer the correct routine to select"""
    objective = dspy.InputField(desc = "The long-term objective (what you are doing)")
    context = dspy.InputField(desc = "The previous actions (what you have done)")
    purpose = dspy.InputField(desc = "The purpose of the action (what you have to do now)")
    prompt = dspy.InputField(desc = "The action specific instructions (How to do it)")
    selected_routine = dspy.OutputField(desc = "The name of the selected routine", prefix="Routine:")

class CallProgramTool(BaseTool):

    def __init__(
            self,
            program_memory: ProgramMemory,
            agent_state: AgentState,
            lm: Optional[dspy.LM] = None,
        ):
        super().__init__(name = "CallProgram", lm = lm)
        self.agent_state = agent_state
        self.program_memory = program_memory
        self.program_name_parser = ProgramNameOutputParser()
        self.prediction_parser = PredictionOutputParser()
        self.predict = dspy.Predict(CallProgramSignature)

    def call_program(self, program_name: str):
        """Method to call a program"""
        if not self.program_memory.exists(program_name):
            return "Error occured: This program does not exist, verify its name"
        if self.program_memory.is_protected(program_name):
            return "Error occured: Trying to call a protected program"
        current_program = self.agent_state.get_current_program()
        current_node = self.agent_state.get_current_node()
        next_node = self.program_memory.get_next_node(current_node, current_program)
        self.agent_state.set_current_node(next_node)
        called_program = self.program_memory.get_graph(program_name)
        first_node = self.program_memory.get_starting_node(program_name)
        self.agent_state.call_program(first_node, called_program)
        return "Successfully called"
        
    def forward(
            self,
            context: str,
            objective: str,
            purpose: str,
            prompt: str,
            disable_inference: bool = False,
        ) -> dspy.Prediction:
        """Method to perform DSPy forward prediction"""
        if not disable_inference:
            with dspy.context(lm=self.lm if self.lm is not None else dspy.settings.lm):
                pred = self.predict(
                    context = context,
                    objective = objective,
                    purpose = purpose,
                    prompt = prompt,
                )
            pred.selected_routine = self.prediction_parser.parse(
                pred.selected_routine, prefix="Routine:", stop=["\n", " "]
            )
            pred.selected_routine = self.program_name_parser.parse(pred.selected_routine)
            observation = self.call_program(pred.selected_routine)
            return dspy.Prediction(
                selected_routine = pred.selected_routine,
                observation = observation,
            )
        else:
            observation = self.call_program(prompt)
            return dspy.Prediction(
                selected_routine = prompt,
                observation = observation,
            )

    def __deepcopy__(self, memo):
        cpy = (type)(self)(
            program_memory = self.program_memory,
            agent_state = self.agent_state,
            lm = self.lm,
        )
        cpy.predict = copy.deepcopy(self.predict)
        return cpy